(* ============================================================== *)
(*                                                                *)
(*                        Z-Transform Formalization               *)
(*                                                                *)
(*        (c) Copyright,                                          *)
(*                       U. Siddique, 2013-2018                   *)
(*                       M. Y. Mahmoud, 2013                      *)
(*                       Hardware Verification Group,             *)
(*                       Concordia University                     *)
(*                                                                *)
(*     Contact: <umair.hvg@gmail.com>                             *)
(*					                                              *)
(*                                                                *)
(*                                                                *)
(*                                                                *)
(*                                                                *)
(* ============================================================== *)


(*===============================================================

let show_types ()  = 
    let print_typed_var fmt tm =
      let s,ty = dest_var tm in
      pp_print_string fmt ("("^s^":"^string_of_type ty^")") in
    install_user_printer("print_typed_var",print_typed_var);;

let hide_types () = delete_user_printer "print_typed_var";;

=================================================================*)


needs "Multivariate/make_complex.ml";;



let POP_SPEC_TAC tm s  = UNDISCH_TAC tm THEN DISCH_TAC THEN 
                      POP_ASSUM (MP_TAC o SPEC s) THEN DISCH_TAC;; 

let POP_AND_SIMP L = REPEAT(POP_ASSUM MP_TAC) THEN SIMP_TAC L THEN 
                   REPEAT(STRIP_TAC);;

let j = define ` j = ii`;;


let z_transform  = new_definition `
    z_transform (f:num->complex) (z:complex) = 
    infsum (from 0) (\n. f(n) / z pow n)`;;


let Z_TRANSFORM = prove (`
	!f z. z_transform f z = 
		  infsum (from 0) (\n. f n * inv z pow n)`,

SIMP_TAC[complex_div;z_transform;GSYM COMPLEX_POW_INV]);;


let z_tr_summable =  new_definition `
    z_tr_summable (f:num->complex) (z:complex) <=> 
                  summable (from 0) (\n. f(n) / z pow n)`;;

let roc = new_definition `
          roc (f:num->complex) = 
          {z | ~(z = Cx(&0))/\ summable (from 0) (\n. f n / z pow n)} `;;




let INFSUM_LIM = prove (`
     !s f.  infsum s f = lim sequentially (\k. vsum (s INTER (0..k)) f)`,

REWRITE_TAC[infsum;sums;GSYM lim]);;


let CPOW_POW = prove (
   ` !n w. ~(w = Cx(&0)) ==> w cpow Cx(&n) = w pow n`,

INDUCT_TAC THENL[
GEN_TAC THEN 
SIMP_TAC[cpow;complex_pow;COMPLEX_MUL_LZERO;CEXP_0];

SIMP_TAC[cpow;complex_pow;COMPLEX_MUL_LZERO;CEXP_0; ADD1;
       CX_ADD;GSYM REAL_OF_NUM_ADD;COMPLEX_ADD_RDISTRIB;
      COMPLEX_MUL_LID;GSYM CEXP_ADD] THEN 
REWRITE_TAC[MESON[GSYM CEXP_ADD]`cexp ((Cx (&n) * clog w) + clog w) 
            = cexp (Cx (&n) * clog w)* cexp(clog w)`] THEN
SIMP_TAC[CEXP_CLOG; 
           COMPLEX_FIELD `w * w pow n =  (w pow n) * w`] THEN 
SIMP_TAC[COMPLEX_EQ_MUL_RCANCEL] THEN 
POP_ASSUM MP_TAC THEN 
SIMP_TAC[cpow]]);;

let ROC_NOT_ZERO = prove(
`!f z. z IN roc f ==> ~(z=Cx(&0))`,
SIMP_TAC[roc;IN_ELIM_THM]);;


(*===================================================================*)
(*   Here we define variant of z_transform using "cpow" and then we  *)
(*   will prove the equivalence. 	   	 	    	                 *)
(*   NOTE: This is quite handy to prove further properties becasue   *)
(*   considerig Z^(n:num) is sometime difficult than z^(Cx(&n))      *)
(*===================================================================*)


let z_tr  = new_definition `
     z_tr (f:num->complex) (z:complex) = 
                  infsum (from 0) (\n. f(n)*z cpow --Cx(&n))`;;


let z_tr_summable_cpow =  new_definition `
    z_tr_summable_cpow (f:num->complex) (z:complex) <=> 
                  summable (from 0) (\n. f(n)*z cpow --Cx(&n))`;;



let ROC = define `
   !f. ROC (f:num->complex) = 
          {z | summable (from 0) (\n. f n * z cpow --Cx (&n))} `;;



let ROC_Z_SUMMABLE_CPOW = prove (` 
    !f z.  z IN ROC f <=> z_tr_summable_cpow f z`,

REWRITE_TAC[IN;ROC;z_tr_summable_cpow;IN_ELIM_THM]);;



let  ROC_COMPLEX_LMUL_CPOW = prove (
    `!a f z. z IN ROC f ==> z IN ROC (\n. a* f n)`,

REWRITE_TAC[IN;ROC;z_tr_summable_cpow;IN_ELIM_THM ] THEN 
REWRITE_TAC[SUMMABLE_COMPLEX_LMUL;GSYM COMPLEX_MUL_ASSOC ]);;


let ROC_LINEAR = prove (`!z a b f g. 
                 z IN ROC f /\ z IN ROC g ==> 
                 z IN ROC (\n. a*f n) INTER ROC (\n.b* g n)`,

SIMP_TAC[INTER;ROC;IN;IN_ELIM_THM;SUMMABLE_COMPLEX_LMUL] THEN 
SIMP_TAC[GSYM COMPLEX_MUL_ASSOC;SUMMABLE_COMPLEX_LMUL]);;


let ROC_SCALING = prove (`!z a f. 
                  z IN ROC f ==> z IN ROC (\n. f n / a)`,

SIMP_TAC[INTER;ROC;IN;IN_ELIM_THM;SUMMABLE_COMPLEX_DIV;
	 COMPLEX_FIELD ` (a/b)*c = (a*c)/b`]);;



let INFSUM_REINDEX = prove
   (`!f n k.  summable  (from n) (\x. f (x + k)) ==>
       infsum  (from (n+k)) f = infsum  (from n) (\x. f(x+k)) `,
   REPEAT STRIP_TAC THEN MATCH_MP_TAC INFSUM_UNIQUE 
   THEN ASM_SIMP_TAC[GSYM SUMS_REINDEX;SUMS_INFSUM]);;


let SUMMABLE_OFFSET = prove 
 (`!f s n m. summable (from m)  f  /\  0 < n/\ m <= n ==> 
    summable (from n)  f`, MESON_TAC[LE_LT;summable;SUMS_OFFSET]);;

let SPEC_V (x,v) thm = (SPEC v o GEN x  o SPEC_ALL) thm;;


let   INFSUM_REINDEX_0 = prove
   (`!f k.  0 < k /\ summable  (from 0) f ==>
       infsum  (from k) f = infsum  (from 0) (\x. f(x+k)) `,
    MESON_TAC[REWRITE_RULE [ADD] (SPEC_V (`n:num`,`0`) INFSUM_REINDEX);
	SUMMABLE_REINDEX;ADD;LE_0;SUMMABLE_OFFSET]);;
		
let SUMMABLE_COMPLEX_LMUL = prove
  (`!c f s.  summable s f ==> summable s (\n. c*(f n)) `, 
   MESON_TAC[summable;SERIES_COMPLEX_LMUL]);;
   
let INFSUM_COMPLEX_LMUL = prove
  (`!c f s.  summable  s f ==> 
    infsum s (\n. c*(f n))  = c * infsum s f`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC INFSUM_UNIQUE 
   THEN ASM_SIMP_TAC[GSYM SERIES_COMPLEX_LMUL;SUMS_INFSUM]);;
  
let INFSUM_OFFSET = prove
(`!f n m. summable  (from m) f 
   /\ 0 < n /\ m <= n ==> 
 infsum (from m)  f =
   vsum (m..n-1) f + infsum  (from n) f`,
 REPEAT GEN_TAC THEN REWRITE_TAC[GSYM SUMS_INFSUM] THEN
 DISCH_THEN(MP_TAC o MATCH_MP SUMS_OFFSET) THEN 
 MESON_TAC[VECTOR_ARITH `(x:complex) = y + z <=> z = x - y`;INFSUM_UNIQUE]);;
 
let TIME_ADVANCE_CPOW  = prove
(`!f z k.  z_tr_summable_cpow f z /\ 0 < k ==> z_tr (\n. f(n+k)) z =
  z cpow Cx (&k)*(z_tr f z - vsum (0..k-1) (\n. f(n)*z cpow --Cx(&n)))`,  
  REPEAT GEN_TAC THEN REWRITE_TAC[z_tr;z_tr_summable_cpow] THEN
  SUBGOAL_THEN `!n. f (n + k) * z cpow --Cx (&n) =  
  z cpow Cx(&k) * f (n + k) * z cpow --Cx (& (n+k) )` ASSUME_TAC
   THENL [ASM_CASES_TAC `z = Cx(&0)`; ALL_TAC]
THENL[ASM_SIMP_TAC[cpow;COMPLEX_MUL_RZERO];REWRITE_TAC[CPOW_ADD; 
GSYM REAL_OF_NUM_ADD;CX_ADD;COMPLEX_NEG_ADD;
COMPLEX_FIELD `x:complex * y * z * e = (x*e) * y * z`]
THEN ASM_SIMP_TAC[ COMPLEX_MUL_RINV;CPOW_NEG;COMPLEX_MUL_LID;CPOW_EQ_0];
ASM_SIMP_TAC[]] THEN IMP_REWRITE_TAC[REWRITE_RULE[] (SPECL [`z cpow Cx (&k) `;
`\n.f (n + k) * z cpow --Cx (&(n + k))`;`from 0`] INFSUM_COMPLEX_LMUL)]
THEN ASM_SIMP_TAC[SPEC_V(`n:num`,`k:num`) INFSUM_OFFSET;LE_0]
THEN ASM_SIMP_TAC[INFSUM_REINDEX_0;COMPLEX_ADD_SUB]
  THEN MESON_TAC[REWRITE_RULE [ADD](ISPECL[`k:num`;
  `(\n. f n * z cpow --Cx (&n))`;`0`] SUMMABLE_REINDEX);SUMMABLE_OFFSET;LE_0]);;
  

let VSUM_OFFSET_0_NEG = prove
(`!f m n. m <= n ==> vsum(m..n) (\i.f (i-m)) = vsum(0..n - m) f`,
SIMP_TAC[VSUM_OFFSET_0;ADD_SUB]);;
 
let SERIES_RESTRICT_NUMSEG = prove
 (`!f k l:real^N.
        ((\n. if k <= n then f(n-k) else vec 0) sums l) (from 0) <=>
        ((\n.f(n-k)) sums l) (from k)`,
		 REWRITE_TAC[FROM_0] THEN
		 REWRITE_TAC[SPEC_V (`k:num->bool`,`from k`) (GSYM SERIES_RESTRICT)
	;IN;from;IN_ELIM_THM]);;
  
let SERIES_NEG_OFFSET = prove
 (`!f k l:real^N.
        (f sums l) (from 0) ==> ((\n.f(n-k)) sums l) (from k)`,
      	REWRITE_TAC[sums;LIM_SEQUENTIALLY;FROM_INTER_NUMSEG] THEN
	IMP_REWRITE_TAC[VSUM_OFFSET_0_NEG] THEN
	 MESON_TAC[ ARITH_RULE `N + k <= n ==> N <= n - k:num`
	 ;ARITH_RULE `N + k <= n ==> k:num <= n `]);;
        
	
let INFSUM_NEG_OFFSET = prove
 (`!f k. summable (from 0) f==> 
        infsum  (from 0) (\n. if k <= n then f(n-k) else vec 0) 
		   = infsum  (from 0) f`,
       REPEAT STRIP_TAC THEN MATCH_MP_TAC INFSUM_UNIQUE THEN
        REWRITE_TAC[SERIES_RESTRICT_NUMSEG] THEN
	MATCH_MP_TAC SERIES_NEG_OFFSET THEN 
	ASM_SIMP_TAC[SUMS_INFSUM]);;
	
let RIGHT_SHIFT_LEMMA = prove
 (`!f z k.  z_tr_summable_cpow f z  ==>
    z_tr (\n. if  k <= n then f(n-k) else Cx(&0)) z = 
	   z cpow --Cx(&k)* z_tr f z `,
	REPEAT GEN_TAC THEN REWRITE_TAC[z_tr_summable_cpow;z_tr] THEN
   SUBGOAL_THEN `!n. k <= n ==> f (n - k) * z cpow --Cx (&n) =  
   z cpow --Cx(&k) * f (n - k) * z cpow --Cx (& (n-k) )` ASSUME_TAC
   THENL[ASM_CASES_TAC `z = Cx(&0)`; ALL_TAC]
   THENL[ASM_SIMP_TAC[cpow;COMPLEX_MUL_RZERO];SIMP_TAC[CPOW_SUB;
    GSYM REAL_OF_NUM_SUB;CX_SUB;COMPLEX_NEG_SUB;
	COMPLEX_FIELD `x:complex * y * z / e = (z*x) * y / e`]
  THEN  ASM_SIMP_TAC[ COMPLEX_MUL_RINV;CPOW_NEG;COMPLEX_MUL_LID;
  CPOW_EQ_0;complex_div];
  ASM_SIMP_TAC[COMPLEX_FIELD`(if p then x else y) * z = if p then x*z else y*z`
   ;COMPLEX_MUL_LZERO;GSYM COMPLEX_VEC_0
   ;INFSUM_NEG_OFFSET;GSYM INFSUM_COMPLEX_LMUL;SUMMABLE_COMPLEX_LMUL]]);;
  

let is_causal_signal = define ` 
    is_causal_signal (f:num->complex) = 
   (!n k.f (n-k)  =  if k <= n then f (n-k) else Cx(&0))`;;



let TIME_DELAY_CPOW = prove(` !f z m.  
                z IN ROC f /\ (is_causal_signal f)  ==>
      	        z_tr (\n. f(n-m))  z = z cpow --Cx(&m)* z_tr f z `,
  
  REWRITE_TAC[is_causal_signal;ROC_Z_SUMMABLE_CPOW] THEN 
  REPEAT STRIP_TAC THEN ONCE_ASM_REWRITE_TAC[] THEN 
  ASM_SIMP_TAC[RIGHT_SHIFT_LEMMA]);;




let ZTR_ADD = prove (` !(f:num->complex)  (g:num->complex) (z:complex). 
 z_tr_summable_cpow f z /\ z_tr_summable_cpow g z ==>
           (z_tr ((\x. f x + g x))  z = z_tr f z +  z_tr g z)`, 

REWRITE_TAC[z_tr;z_tr_summable_cpow;INFSUM_ADD;COMPLEX_ADD_RDISTRIB]);;

let ZTR_SUB = prove (` !(f:num->complex)  (g:num->complex) (z:complex). 
 z_tr_summable_cpow f z /\ z_tr_summable_cpow g z ==>
           (z_tr ((\x. f x - g x))  z = z_tr f z -  z_tr g z)`, 

REWRITE_TAC[z_tr;z_tr_summable_cpow;INFSUM_SUB;COMPLEX_SUB_RDISTRIB]);;


let  ZTR_CONST_MUL = prove (` ! c f z. z_tr_summable_cpow f z ==> 
        z_tr (\n. c* f n) z = c * z_tr (\n. f n) z`,

REWRITE_TAC[z_tr;z_tr_summable_cpow;GSYM COMPLEX_MUL_ASSOC;INFSUM_COMPLEX_LMUL]);;


let ZTR_LINEAR_ALT = prove (` !(f:num->complex)  (g:num->complex) (z:complex) a b. 
 z_tr_summable_cpow f z /\ z_tr_summable_cpow g z ==>
           (z_tr ((\x. a*f x + b*g x))  z = a*z_tr f z +  b*z_tr g z)`, 

REWRITE_TAC[z_tr;z_tr_summable_cpow;GSYM COMPLEX_MUL_ASSOC;INFSUM_COMPLEX_LMUL;
ZTR_ADD;ZTR_CONST_MUL;INFSUM_ADD;COMPLEX_ADD_RDISTRIB;SUMMABLE_COMPLEX_LMUL] THEN 
REPEAT STRIP_TAC THEN 
ONCE_SIMP_TAC[ COMPLEX_FIELD 
 `(\n. a * f n * z cpow --Cx (&n) + b * g n * z cpow --Cx (&n)) = 
(\n. (\n. a * f n * z cpow --Cx (&n)) n + (\n. b * g n * z cpow --Cx (&n)) n)`] 
THEN IMP_REWRITE_TAC [INFSUM_ADD] THEN 
ASM_SIMP_TAC[SUMMABLE_COMPLEX_LMUL]
 THEN ASM_SIMP_TAC[INFSUM_COMPLEX_LMUL]);;


let ZTR_LINEAR = prove (` !(f:num->complex)  (g:num->complex) (z:complex) a b z. 
 z IN ROC f INTER  ROC g  ==>
           (z_tr ((\x. a*f x + b*g x))  z = a*z_tr f z +  b*z_tr g z)`, 

SIMP_TAC[ROC_Z_SUMMABLE_CPOW;ZTR_LINEAR_ALT;INTER;IN;IN_ELIM_THM]);;



let Z_SUMMABLE_CMUL_DIFF = prove (` 
      !c f k z. z_tr_summable_cpow f z /\ is_causal_signal f
                          ==> z_tr_summable_cpow (\x. c * f ( x - k )) z`, 

REWRITE_TAC[z_tr_summable_cpow;is_causal_signal] THEN 
REPEAT STRIP_TAC THEN 
SIMP_TAC[GSYM COMPLEX_MUL_ASSOC] THEN 
MATCH_MP_TAC SUMMABLE_COMPLEX_LMUL THEN 
ONCE_ASM_REWRITE_TAC[] THEN 
ASSERT_TAC `!n. k <= n ==> f (n - k) * z cpow --Cx (&n) =  
   z cpow --Cx(&k) * f (n - k) * z cpow --Cx (& (n-k) ) ` THENL
   [ASM_CASES_TAC `z = Cx(&0)`; ALL_TAC]
   THENL[ASM_SIMP_TAC[cpow;COMPLEX_MUL_RZERO];

SIMP_TAC[CPOW_SUB;
    GSYM REAL_OF_NUM_SUB;CX_SUB;COMPLEX_NEG_SUB;
	COMPLEX_FIELD `x:complex * y * z / e = (z*x) * y / e`]
  THEN  ASM_SIMP_TAC[ COMPLEX_MUL_RINV;CPOW_NEG;COMPLEX_MUL_LID;
  CPOW_EQ_0;complex_div];
  ASM_SIMP_TAC[COMPLEX_FIELD`(if p then x else y) * z = if p then x*z else y*z`
   ;COMPLEX_MUL_LZERO;GSYM COMPLEX_VEC_0
   ;INFSUM_NEG_OFFSET;GSYM INFSUM_COMPLEX_LMUL;SUMMABLE_COMPLEX_LMUL]] 
THEN 
REPEAT(POP_ASSUM MP_TAC) THEN 
REWRITE_TAC[summable] THEN 
REPEAT STRIP_TAC THEN 
ONCE_REWRITE_TAC[COMPLEX_FIELD `!l. z cpow --Cx (&k) * f (n - k) * z cpow --Cx (&(n - k)) = 
(\n. z cpow --Cx (&k) * f (n) * z cpow --Cx (&(n))) (n - k) `] THEN 
ONCE_REWRITE_TAC[SERIES_RESTRICT_NUMSEG] THEN 
ONCE_REWRITE_TAC[GSYM summable] THEN 
SIMP_TAC[] THEN 
MATCH_MP_TAC SUMMABLE_COMPLEX_LMUL THEN 
REWRITE_TAC[summable] THEN 
EXISTS_TAC`l:complex` THEN 
ONCE_REWRITE_TAC[COMPLEX_FIELD `(\n. f (n - k) * z cpow --Cx (&(n - k)))= 
                (\n. (\n. f (n) * z cpow --Cx (&(n)))(n-k))`] THEN 
MATCH_MP_TAC SERIES_NEG_OFFSET THEN 
ASM_SIMP_TAC[]);;

let Z_SUMMABLE_LINEAR = prove (
   `! f g z.  z_tr_summable_cpow f z /\ z_tr_summable_cpow g z ==>
                              z_tr_summable_cpow (\x. f x + g x) z`, 

REWRITE_TAC[z_tr_summable_cpow] THEN 
REWRITE_TAC[COMPLEX_ADD_RDISTRIB] THEN
ASM_SIMP_TAC[SUMMABLE_ADD]);;


let FIRST_DIFFERENCE = prove(`!f z. 
    z IN ROC f /\ (is_causal_signal f) ==>
    z_tr (\n. f n - f(n-1)) z = (Cx(&1) - z cpow Cx(-- &1))* z_tr f z`,

REWRITE_TAC[ROC_Z_SUMMABLE_CPOW] THEN REPEAT STRIP_TAC THEN
ONCE_REWRITE_TAC[COMPLEX_FIELD`(\n. f n - f(n-1)) = 
                          (\n. f n - (\n. f(n-1)) n)`] THEN
IMP_REWRITE_TAC [ZTR_SUB] THEN 
ONCE_REWRITE_TAC[COMPLEX_FIELD ` !n. f (n - 1) = Cx(&1)* f (n - 1)`] THEN 
IMP_REWRITE_TAC[Z_SUMMABLE_CMUL_DIFF] THEN 
SIMP_TAC[COMPLEX_MUL_LID] THEN IMP_REWRITE_TAC[TIME_DELAY_CPOW] THEN 
ASM_SIMP_TAC[ROC_Z_SUMMABLE_CPOW] THEN SIMP_TAC[CX_NEG] THEN 
CONV_TAC COMPLEX_FIELD);;



let INV_CPOW = prove (
   `!n z. ~(z = Cx(&0)) ==>  z cpow --Cx(&n) = (inv z) cpow Cx(&n)`,  

INDUCT_TAC THENL[
REWRITE_TAC[CPOW_NEG] THEN 
SIMP_TAC[cpow; COMPLEX_FIELD ` !z. 
                     ~(z = Cx (&0)) ==> ~(inv z = Cx (&0))`] THEN 
SIMP_TAC[COMPLEX_MUL_LZERO;CEXP_0] THEN 
CONV_TAC COMPLEX_FIELD;

REWRITE_TAC[ADD1;GSYM REAL_OF_NUM_ADD;CX_ADD] THEN 
SIMP_TAC[COMPLEX_FIELD 
         `--(Cx (&n) + Cx (&1))= (--Cx (&n) + --Cx (&1)) `;
         GSYM CX_NEG; CPOW_ADD] THEN 
REWRITE_TAC[CX_NEG] THEN 
ASM_SIMP_TAC[] THEN 
SIMP_TAC[COMPLEX_EQ_MUL_LCANCEL] THEN 
REPEAT STRIP_TAC THEN 
DISJ2_TAC THEN
ASM_SIMP_TAC[cpow; COMPLEX_FIELD ` !z. 
                     ~(z = Cx (&0)) ==> ~(inv z = Cx (&0))`] THEN 
SIMP_TAC[COMPLEX_MUL_LID; 
        COMPLEX_FIELD `--Cx (&1) * clog z= --(Cx (&1) * clog z)`] THEN 
ASM_SIMP_TAC[CEXP_CLOG;CEXP_NEG; COMPLEX_FIELD ` !z. 
                     ~(z = Cx (&0)) ==> ~(inv z = Cx (&0))`]]);;



let CPOW_INV = prove ( ` 
      !a k. (Im a = &0 ==> &0 < Re a) ==> a cpow k = inv a cpow --k`,

REPEAT GEN_TAC THEN ASM_CASES_TAC `a = Cx(&0)` THEN 
ASM_SIMP_TAC[cpow;COMPLEX_MUL_RZERO;COMPLEX_INV_0] THEN 
ASM_SIMP_TAC[cpow;COMPLEX_MUL_RZERO;COMPLEX_INV_0; 
COMPLEX_FIELD ` ~(a = Cx(&0)) ==> ~(inv a = Cx(&0))`;
COMPLEX_FIELD ` --a*b = --(a*b)`;CLOG_INV;
COMPLEX_FIELD `--(k * --clog a) = k * clog a`]);;



let SCALING_IN_Z_DOAMIN = prove (` !f z a. 
     z_tr (\n. a cpow Cx(&n) * f n) z = z_tr (\n. f n) (inv a*z)`,

REPEAT GEN_TAC THEN 
ASM_CASES_TAC `z = Cx(&0)` THENL[ 
ASM_SIMP_TAC[z_tr; cpow;COMPLEX_MUL_LZERO;COMPLEX_MUL_RZERO];
ALL_TAC] THEN 
ASM_CASES_TAC `a = Cx(&0)` THENL[
ASM_SIMP_TAC[z_tr; cpow;COMPLEX_MUL_LZERO;COMPLEX_MUL_RZERO;
             COMPLEX_INV_0]; ALL_TAC] THEN 
REPEAT (POP_ASSUM MP_TAC) THEN 
SIMP_TAC[z_tr;INV_CPOW] THEN 
SIMP_TAC[INV_CPOW;COMPLEX_INV_INV;COMPLEX_POW_MUL;
         COMPLEX_FIELD `~(z = Cx (&0)) /\ ~(a = Cx (&0))
         ==>  ~(inv a * z = Cx(&0))`;COMPLEX_INV_MUL;
         COMPLEX_FIELD `~(z = Cx (&0)) /\ ~(a = Cx (&0))
         ==>  ~(a * inv z = Cx(&0))`;
	 COMPLEX_FIELD `~(z = Cx (&0)) 
         ==>  ~(inv z = Cx(&0))` ;CPOW_POW] THEN 
SIMP_TAC[COMPLEX_FIELD `!n. (a pow n * f n) * inv z pow n = 
                        f n * a pow n * inv z pow n`]);;


let SCALING_IN_Z_DOAMIN_NEG = prove (` !f z w. 
    z_tr (\n. w cpow --Cx(&n) * f n) z = z_tr (\n. f n) (w*z)`,

REPEAT GEN_TAC THEN 
ASM_CASES_TAC `z = Cx(&0)` THENL[ 
ASM_SIMP_TAC[z_tr; cpow;COMPLEX_MUL_LZERO;COMPLEX_MUL_RZERO];
ALL_TAC] THEN 
ASM_CASES_TAC `w = Cx(&0)` THENL[
ASM_SIMP_TAC[z_tr; cpow;COMPLEX_MUL_LZERO;COMPLEX_MUL_RZERO;
             COMPLEX_INV_0]; ALL_TAC] THEN 
REPEAT (POP_ASSUM MP_TAC) THEN 
SIMP_TAC[z_tr;INV_CPOW] THEN 
SIMP_TAC[INV_CPOW;COMPLEX_INV_INV;COMPLEX_POW_MUL;
         COMPLEX_FIELD `~(z = Cx (&0)) /\ ~(w = Cx (&0))
         ==>  ~(w * z = Cx(&0))`;COMPLEX_INV_MUL;
         COMPLEX_FIELD `~(z = Cx (&0)) /\ ~(a = Cx (&0))
         ==>  ~(a * inv z = Cx(&0))`;
	 COMPLEX_FIELD `~(z = Cx (&0)) 
         ==>  ~(inv z = Cx(&0))` ;CPOW_POW] THEN 
SIMP_TAC[COMPLEX_FIELD `!n. (a pow n * f n) * inv z pow n = 
                        f n * a pow n * inv z pow n`]);;



(*---------Special Cases when "a" and "w" are REALS-------*)


let SCALING_Z = prove (`!a z f. 
real a /\ real z /\ &0 < Re (a) /\ &0 <= Re z ==>            
(z_tr (\n. a cpow Cx(&n) * f n) z = z_tr (\n. f n) (inv a*z))`,
 
ONCE_REWRITE_TAC[FORALL_COMPLEX] THEN 
REPEAT GEN_TAC THEN REPEAT STRIP_TAC THEN ASM_CASES_TAC `z = Cx(&0)` THENL[
ASM_SIMP_TAC[z_tr;COMPLEX_FIELD `! a b c. (a*b)*c = b*(a*c)`;cpow;
COMPLEX_MUL_RZERO;COMPLEX_INV_0];ALL_TAC] THEN 
ASM_SIMP_TAC[z_tr;COMPLEX_FIELD `! a b c. (a*b)*c = b*(a*c)`;GSYM CEXP_ADD] 
THEN ASSERT_TAC `!n. complex (x,y) cpow Cx (&n) = 
                 inv (complex (x,y)) cpow --Cx (&n)` THENL[

GEN_TAC THEN MATCH_MP_TAC CPOW_INV THEN ASM_SIMP_TAC[];ALL_TAC] THEN 
ASM_SIMP_TAC[] THEN 
IMP_REWRITE_TAC [(GSYM CPOW_MUL_REAL)] THEN 
ASM_SIMP_TAC[REAL_INV] THEN 
SIMP_TAC[complex_inv;RE;IM] THEN 
IMP_REWRITE_TAC [REAL_LE_DIV] THEN 
ASM_SIMP_TAC[REAL_FIELD `&0 <= x pow 2 + y pow 2 <=> 
                         &0 + &0 <= x pow 2 + y pow 2`; 
REAL_LE_ADD2;REAL_LE_LT;REAL_LE_POW_2] THEN 
ASM_MESON_TAC[RE]);;




let COMPLEX_SCALE_CHANGE = prove(` ! w z f. 
    real w /\ real z /\ &0 <= Re w /\ &0 <= Re z ==>
    z_tr (\n. w cpow --Cx(&n) * f n) z =  z_tr f (w*z)`,

REWRITE_TAC[z_tr] THEN 
SIMP_TAC[COMPLEX_FIELD `(w cpow --Cx (&n) * f n) * z cpow --Cx (&n) = 
                       f n * (w cpow --Cx (&n) * z cpow --Cx (&n)) `] THEN
SIMP_TAC[ CPOW_MUL_REAL]);;



let COMPLEX_DERIV_CPOW = prove (`!s. &0 < Re z ==> 
  complex_derivative (\z. z cpow s) z = s * z cpow (s - Cx (&1))`,

GEN_TAC THEN  DISCH_TAC THEN MATCH_MP_TAC HAS_COMPLEX_DERIVATIVE_DERIVATIVE 
THEN ASM_SIMP_TAC[HAS_COMPLEX_DERIVATIVE_CPOW]);;


(*===================================================================*)
(* Now we can use above theorms to prove main properties 
                     of z_transform                                  *)
(*===================================================================*)


let ZTR_EQ = prove(
`!z f.  ~(z =Cx(&0)) ==>  z_transform (f:num->complex) z = z_tr f z`,
SIMP_TAC[z_transform;z_tr;CPOW_POW;CPOW_NEG;complex_div]);;


let ZTR_ROC_EQ = prove(
`!z f.  z IN roc f ==> z_tr_summable_cpow f z`,
SIMP_TAC[z_tr_summable_cpow;roc;IN_ELIM_THM;CPOW_POW;CPOW_NEG;complex_div]);;

let ZTR_ROC_ROC = prove(
`!z f.  z IN roc f ==>  z IN ROC f `,
SIMP_TAC[ROC;roc;IN_ELIM_THM;CPOW_POW;CPOW_NEG;complex_div]);;


let ZTR_ZTRANSFORM_TAC L = 
    IMP_REWRITE_TAC (L@[ZTR_EQ;ZTR_ROC_EQ;ZTR_ROC_ROC;ROC_NOT_ZERO]);;



let Z_TRANSFORM_ADD = prove(
        `!f g z.
         z IN roc f /\ z IN roc  g   
        ==> z_transform (\x. f x + g x) z = 
             z_transform f z + z_transform g z`,

REPEAT STRIP_TAC THEN 
ZTR_ZTRANSFORM_TAC[ZTR_ADD]);;


let Z_TRANSFORM_SUB = prove(
        `!f g z.
         z IN roc f /\ z IN roc  g   
        ==> z_transform (\x. f x - g x) z = 
             z_transform f z - z_transform g z`,

REPEAT STRIP_TAC THEN 
ZTR_ZTRANSFORM_TAC[ZTR_SUB]);;

let Z_TRANSFORM_CONST_MUL = prove(
        `!c f z.
         z IN roc f
         ==> z_transform (\n. c * f n) z = c * z_transform (\n. f n) z`,

REPEAT STRIP_TAC THEN 
ZTR_ZTRANSFORM_TAC[ZTR_CONST_MUL ]);;

let Z_TRANSFORM_LINEAR_ALT  = prove(
        `!f g z a b.
         z IN roc f /\ z IN roc g
         ==> z_transform (\x. a * f x + b * g x) z = 
             a * z_transform f z + b * z_transform g z`,

REPEAT STRIP_TAC THEN 
ZTR_ZTRANSFORM_TAC[ZTR_LINEAR_ALT ]);;

let Z_TRANSFORM_LINEAR  = prove(
        `!f g z a b.
         z IN roc f  INTER roc g
         ==> z_transform (\x. a * f x + b * g x) z = 
             a * z_transform f z + b * z_transform g z`,

SIMP_TAC[IN_INTER; Z_TRANSFORM_LINEAR_ALT]);;


let Z_TRANSFORM_FIRST_DIFFERENCE  = prove(
        `!f z.
         z IN roc f /\ is_causal_signal f
         ==> z_transform (\n. f n - f (n - 1)) z =
             (Cx (&1) - z cpow Cx (-- &1)) * z_transform f z`,

REPEAT STRIP_TAC THEN 
ZTR_ZTRANSFORM_TAC[FIRST_DIFFERENCE]);;

let Z_TRANSFORM_SCALING_IN_Z  = prove(
        `!f z a. inv a*z IN roc f /\ z IN roc f  ==>
                 z_transform (\n. a cpow Cx (&n) * f n) z = 
                 z_transform (\n. f n) (inv a * z)`,

REPEAT STRIP_TAC THEN 
ZTR_ZTRANSFORM_TAC[SCALING_IN_Z_DOAMIN]);;


let Z_TRANSFORM_SCALING_IN_Z_NEG  = prove(
        `!f z a. a*z IN roc f /\ z IN roc f  ==>
                 z_transform (\n. a cpow --Cx (&n) * f n) z = 
                 z_transform (\n. f n) (a * z)`,

REPEAT STRIP_TAC THEN 
ZTR_ZTRANSFORM_TAC[SCALING_IN_Z_DOAMIN_NEG]);;


let Z_TRANSFORM_TIME_ADVANCE = prove(
`!f z k.  z IN roc f /\ 0 < k ==> z_transform (\n. f(n+k)) z =
  z pow k * (z_transform f z - vsum (0..k-1) (\n. f(n) / z pow n))`,
IMP_REWRITE_TAC[ZTR_EQ;TIME_ADVANCE_CPOW;ZTR_ROC_EQ;ROC_NOT_ZERO;CPOW_POW;
                CPOW_NEG;complex_div]);;


let ZTRANSFORM_TIME_DELAY = prove(` !f z m.  
                z IN roc f /\ (is_causal_signal f)  ==>
      	        z_transform (\n. f(n-m))  z =   z_transform f z / z pow m`,
IMP_REWRITE_TAC[ZTR_EQ;TIME_DELAY_CPOW;REWRITE_RULE[ROC] ZTR_ROC_ROC;ROC_NOT_ZERO;
CPOW_POW;CPOW_NEG;complex_div; COMPLEX_MUL_SYM] THEN SIMP_TAC[ZTR_ROC_ROC]);;



let CEXP_1 = prove (
 ` !z.  ~(z = Cx (&0))  ==> z cpow Cx(&1) = z`,

SIMP_TAC[cpow;COMPLEX_MUL_LID;CEXP_CLOG]);;


let COMPLEX_DERIV_VSUM = prove(` 
!f s. FINITE s /\ (!a. a IN s ==> (f a has_complex_derivative f' a) (at z))
   ==>  complex_derivative  (\x. vsum s (\a. f a x)) z = vsum s f'`,

REPEAT GEN_TAC THEN REPEAT STRIP_TAC THEN  
MATCH_MP_TAC HAS_COMPLEX_DERIVATIVE_DERIVATIVE THEN 
MATCH_MP_TAC HAS_COMPLEX_DERIVATIVE_VSUM THEN 
ASM_SIMP_TAC[]);;


let Z_DERIV_LEMMA = prove (` ! f z. ~(z = Cx (&0)) /\ &0 < Re z /\
      z_tr_summable_cpow  (\n. inv z * Cx(&n)* f n) z
      ==>  z_tr (\n. Cx(&n)* f n) z = 
z * infsum (from 0) (\n. complex_derivative (\z. --f n * z cpow Cx (-- &n)) z)`,

REWRITE_TAC[z_tr;z_tr_summable_cpow;GSYM CX_NEG] THEN
ASSERT_TAC `!n. Cx(-- &n) = Cx((-- &n - &1) + (&1))` THENL[
  REWRITE_TAC[CX_INJ] THEN CONV_TAC REAL_FIELD;ALL_TAC]THEN 
ASM_REWRITE_TAC[] THEN 
SIMP_TAC[CX_ADD; CPOW_ADD;CEXP_1;MESON[] `! a b c. a /\ b  ==> c <=> a ==> b  ==> c`] THEN 
REWRITE_TAC[COMPLEX_FIELD ` !n z. (Cx (&n) * f n) * z cpow Cx (-- &n - &1) * z  = 
                   z*((Cx (&n) * f n) * z cpow Cx (-- &n - &1))`] THEN 
ONCE_REWRITE_TAC[ COMPLEX_FIELD ` !n. (inv z * Cx (&n) * f n) * z cpow Cx (-- &n - &1) * z = 
           (z*inv z) * (Cx (&n) * f n * z cpow Cx (-- &n - &1))` ] THEN 
SIMP_TAC[GSYM complex_div;COMPLEX_DIV_REFL;COMPLEX_MUL_LID] THEN 
REWRITE_TAC[INFSUM_COMPLEX_LMUL;z_tr_summable_cpow;SUMMABLE_COMPLEX_LMUL] THEN 
REPEAT STRIP_TAC THEN 
ONCE_REWRITE_TAC[ COMPLEX_FIELD `(\n. z * (Cx (&n) * f n) * z cpow Cx (-- &n - &1)) =
 (\n. z * (\n. ((Cx (&n) * f n) * z cpow Cx (-- &n - &1))) n)`] 
THEN  IMP_REWRITE_TAC[INFSUM_COMPLEX_LMUL] THEN
ASM_SIMP_TAC[GSYM COMPLEX_MUL_ASSOC]  THEN 
REWRITE_TAC[CX_SUB] THEN 
REWRITE_TAC[COMPLEX_FIELD `!n.   
Cx (&n) * f n * z cpow (Cx (-- &n) - Cx (&1)) = 
 -- (f n) * (-- Cx (&n) * z cpow (Cx (-- &n) - Cx (&1)))`] THEN 
REWRITE_TAC[GSYM CX_NEG] THEN 
IMP_REWRITE_TAC [GSYM COMPLEX_DERIV_CPOW] THEN ASM_SIMP_TAC[] 
THEN ASSERT_TAC `!n. --f n * complex_derivative (\z. z cpow Cx (-- &n)) z = 
complex_derivative (\z. --f n * z cpow Cx(-- &n)) z` THENL[
GEN_TAC THEN ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN
MATCH_MP_TAC  COMPLEX_DERIVATIVE_LMUL THEN 
ONCE_REWRITE_TAC[  complex_differentiable] THEN 
EXISTS_TAC` Cx(-- &n) * z cpow (Cx(-- &n) - Cx (&1))` THEN 
ASM_SIMP_TAC[HAS_COMPLEX_DERIVATIVE_CPOW];ALL_TAC] THEN 
ONCE_ASM_REWRITE_TAC[] THEN 
REWRITE_TAC[GSYM CPOW_ADD;COMPLEX_MUL_LID] THEN 
REWRITE_TAC[GSYM CX_ADD;GSYM CX_SUB] THEN 
SIMP_TAC[REAL_FIELD `((-- &n - &1 + &1) - &1 + &1) = -- &n`]
THEN POP_ASSUM MP_TAC THEN ONCE_SIMP_TAC[EQ_SYM_EQ] THEN 
SIMP_TAC[REAL_FIELD `((-- &n - &1 + &1)) = -- &n`]
);;



let Z_DERIV_ALT_GEN  = prove (`!f z. ~(z = Cx (&0)) /\ &0 < Re z /\ 
       z_tr_summable_cpow (\n. Cx (&n) * f n) z  ==> 
       z_tr (\n. Cx (&n) * f n) z =
             z * infsum (from 0)
             (\n. complex_derivative (\z. --f n * z cpow Cx (-- &n)) z)`,

REPEAT STRIP_TAC THEN 
MATCH_MP_TAC Z_DERIV_LEMMA THEN 
ASM_SIMP_TAC[] THEN 
SIMP_TAC[z_tr_summable_cpow;GSYM COMPLEX_MUL_ASSOC] THEN 
ASSERT_TAC ` (\n. inv z * Cx (&n) * f n * z cpow --Cx (&n)) = 
 (\n. inv z * Cx (&n) * f n * z cpow --Cx (&n))` THENL[
SIMP_TAC[];ALL_TAC] THEN ASM_SIMP_TAC [] THEN 

MATCH_MP_TAC SUMMABLE_COMPLEX_LMUL THEN
ASM_SIMP_TAC[GSYM z_tr_summable_cpow;COMPLEX_MUL_ASSOC]);;


let Z_DERIV  = prove (`!f z. ~(z = Cx (&0)) /\ &0 < Re z /\ 
       z IN ROC (\n. Cx (&n) * f n) ==> 
       z_tr (\n. Cx (&n) * f n) z =
             z * infsum (from 0)
             (\n. complex_derivative (\z. --f n * z cpow Cx (-- &n)) z)`,

MESON_TAC[ROC_Z_SUMMABLE_CPOW;Z_DERIV_ALT_GEN]);;


let CNJ_INFSUM = prove (
    `!f s. summable s f ==>  
    	infsum s (\n. cnj (f n)) = cnj (infsum s f)`,

REPEAT STRIP_TAC THEN MATCH_MP_TAC INFSUM_UNIQUE THEN 
  REWRITE_TAC[SUMS_CNJ] THEN ASM_SIMP_TAC[SUMS_INFSUM]);;



let Z_TRANSFORM_CNJ = prove (` 
!f z. (cnj z) IN roc f ==>
         z_transform (\n. cnj (f n))  z = 
         cnj (z_transform f (cnj z))`,

SIMP_TAC[z_transform;CNJ_DIV] THEN
ONCE_REWRITE_TAC[MESON[CNJ_CNJ]` 
   !z n. z pow n = cnj (cnj (z pow n))`] THEN
ONCE_REWRITE_TAC[GSYM CNJ_DIV] THEN
ONCE_REWRITE_TAC[CNJ_POW] THEN
ONCE_REWRITE_TAC[GSYM CNJ_POW;CNJ_CNJ] THEN
REWRITE_TAC[GSYM CNJ_POW;CNJ_CNJ] THEN
SIMP_TAC[ROC;IN_ELIM_THM] THEN
ONCE_REWRITE_TAC[MESON[CNJ_CNJ]` 
   !z n.  f n / cnj (z pow n) = 
         cnj(cnj(f n))/ cnj (z pow n)`] THEN
REWRITE_TAC[GSYM CNJ_DIV] THEN
ONCE_REWRITE_TAC[GSYM CNJ_POW;CNJ_CNJ] THEN
ONCE_REWRITE_TAC[MESON[CNJ_CNJ]` 
   !z n.  f n / cnj (z pow n) = 
         cnj(cnj(f n))/ cnj (z pow n)`] THEN
REPEAT STRIP_TAC THEN
IMP_REWRITE_TAC[GSYM CNJ_INFSUM] THEN
SIMP_TAC [CNJ_DIV;CNJ_CNJ] THEN
POP_ASSUM MP_TAC THEN
SIMP_TAC[roc;CNJ_DIV;IN_ELIM_THM;CNJ_POW]);;


(*=======================================================================*)
(*           Z-Transform of some common functions (signals)              *)
(*=======================================================================*)

let delta = define ` delta a = (\n. if (n = a) then Cx(&1) else Cx(&0))`;;

let Z_TRANSFORM_DELTA_FUNC_LEMMA = prove
 (`!z.  z_transform (\n. if n = 0 then Cx(&1) else Cx(&0)) z = Cx(&1)`,
  GEN_TAC THEN REWRITE_TAC [z_transform] THEN MATCH_MP_TAC INFSUM_UNIQUE THEN
  ONCE_REWRITE_TAC [sums] THEN
  SIMP_TAC
    [complex_div;
     COMPLEX_FIELD
       `(if b then (c:complex) else d)*k = 
          (if b then (c:complex)*k else d*k)`] THEN
  SIMP_TAC
    [FROM_INTER_NUMSEG; COMPLEX_MUL_LID; COMPLEX_MUL_LZERO; complex_pow;
     COMPLEX_INV_1] THEN
  REWRITE_TAC [GSYM COMPLEX_VEC_0] THEN SIMP_TAC [VSUM_DELTA] THEN
  SIMP_TAC [IN_NUMSEG_0; ARITH_RULE `0 <= (n:num)`] THEN
  SIMP_TAC [sequentially; LIM_CONST]);;


let Z_TRANSFORM_DELTA_FUNC_LEMMA_GEN = prove
 (`!z.  z_transform (\n. if n = a then Cx(&1) else Cx(&0)) z = inv z pow a`,
  GEN_TAC THEN REWRITE_TAC [z_transform] THEN MATCH_MP_TAC INFSUM_UNIQUE THEN
  SIMP_TAC
    [complex_div;
     COMPLEX_FIELD
       `(if b then (c:complex) else d)*k = 
          (if b then (c:complex)*k else d*k)`] THEN
  SIMP_TAC
    [FROM_INTER_NUMSEG; COMPLEX_MUL_LID; COMPLEX_MUL_LZERO; complex_pow;
     COMPLEX_INV_1] THEN
  MATCH_MP_TAC SERIES_VSUM THEN EXISTS_TAC `0..a` THEN
  SIMP_TAC [FINITE_NUMSEG] THEN REPEAT STRIP_TAC THENL
   [SIMP_TAC [SUBSET] THEN SIMP_TAC [IN_FROM; IN] THEN ARITH_TAC;
    POP_ASSUM MP_TAC THEN
    SIMP_TAC [IN_NUMSEG; ARITH_RULE `~(0 <= x /\ x <= a) ==> ~(x = a)`] THEN
    SIMP_TAC [COMPLEX_VEC_0];
    SIMP_TAC [GSYM COMPLEX_VEC_0] THEN SIMP_TAC [VSUM_DELTA] THEN
    SIMP_TAC [IN_NUMSEG_0; ARITH_RULE `a <= (a:num)`] THEN
    SIMP_TAC [COMPLEX_POW_INV]]);;


let Z_TRANSFORM_DELTA_FUNC = prove 
(`!z n. z_transform (delta n) z = inv z pow n`,

SIMP_TAC[delta;Z_TRANSFORM_DELTA_FUNC_LEMMA_GEN]);;



let Z_TRANSFORM_NEG_EXPONETIAL = prove
 (`!z a. inv (norm (cexp a)) / norm z < &1 ==>
z_transform (\n. cexp (--(Cx(&n)*a))) z = 
        inv (cexp a * z) pow 0 / (Cx(&1) -  inv (cexp a * z))`,
  REPEAT STRIP_TAC THEN REWRITE_TAC [z_transform] THEN
  MATCH_MP_TAC INFSUM_UNIQUE THEN SIMP_TAC [CEXP_NEG; CEXP_N] THEN
  SIMP_TAC [GSYM COMPLEX_POW_INV; GSYM COMPLEX_POW_DIV] THEN
  SIMP_TAC [complex_div; GSYM COMPLEX_INV_MUL] THEN
  REWRITE_TAC [GSYM complex_div] THEN MATCH_MP_TAC SUMS_GP THEN
  SIMP_TAC [COMPLEX_INV_MUL; COMPLEX_NORM_MUL] THEN
  ASM_CASES_TAC `z = Cx(&0)` THENL
   [ASM_SIMP_TAC [COMPLEX_INV_0; COMPLEX_NORM_0; REAL_MUL_RZERO] THEN
    CONV_TAC REAL_FIELD;
    SIMP_TAC [COMPLEX_NORM_INV] THEN ASM_SIMP_TAC [GSYM real_div]]);;


let Z_TRANSFORM_NEG_EXPONETIAL_GEN = 
    SIMP_RULE [COMPLEX_FIELD `!n a. (--(Cx (&n) * a)) = (a * --(Cx (&n)))`]
    Z_TRANSFORM_NEG_EXPONETIAL;;



let Z_TRANSFORM_POS_EXPONETIAL_I = prove
 (`!z w . norm (cexp (j * w) * inv z) < &1 ==> 
  z_transform (\n. cexp (j * w * Cx (&n))) z = 
  (cexp (j * w) * inv z) pow 0 / 
  (Cx (&1) - cexp (j * w) * inv z)`,
  REWRITE_TAC[j] THEN 
  REPEAT STRIP_TAC THEN SIMP_TAC [z_transform] THEN
  SIMP_TAC [complex_div; GSYM COMPLEX_POW_INV] THEN
  SIMP_TAC [COMPLEX_FIELD `! w j. (j * w * Cx (&n)) = (Cx (&n) * (j * w))`] THEN
  SIMP_TAC [CEXP_N] THEN SIMP_TAC [GSYM COMPLEX_POW_MUL] THEN
  SIMP_TAC [GSYM complex_div] THEN MATCH_MP_TAC INFSUM_UNIQUE THEN
  IMP_REWRITE_TAC [SUMS_GP] THEN ASM_SIMP_TAC [complex_div]);;

let Z_TRANSFORM_POS_EXPONETIAL = REWRITE_RULE[j]  Z_TRANSFORM_POS_EXPONETIAL_I ;;

let Z_TRANSFORM_CONST_N = prove
 (`!z a. norm (a / z) < &1 ==>
z_transform (\n. a pow n) z = (a / z) pow 0 / (Cx (&1) - (a / z))`,
  REPEAT STRIP_TAC THEN REWRITE_TAC [z_transform] THEN
  MATCH_MP_TAC INFSUM_UNIQUE THEN REWRITE_TAC [GSYM COMPLEX_POW_DIV] THEN
  MATCH_MP_TAC SUMS_GP THEN ASM_SIMP_TAC []);;


let Z_TRANSFORM_CONST_N_ALT =  
REWRITE_RULE [complex_pow] Z_TRANSFORM_CONST_N;;


let S_LEMMA = prove ( `!j. (\n. cexp (j * w * Cx (&n)) * inv (Cx (&2)) +
      cexp (--j * w * Cx (&n)) * inv (Cx (&2))) = 
 (\n.   inv (Cx (&2)) * (\n. cexp (j * w * Cx (&n))) n  +
        inv (Cx (&2)) * (\n. cexp (--j * w * Cx (&n))) n) `,

SIMP_TAC[COMPLEX_MUL_SYM]);;


let CCOS_ALT = 
    SIMP_RULE [COMPLEX_FIELD ` a = b / Cx(&2) <=> Cx(&2)*a = b`]
    ccos;;

let Z_TRANSFORM_CCOS = prove
 (`!(z:complex) (w:real). &1 < (norm z) /\ 
     ~(z = cexp (j * Cx w)) /\
     ~(z = cexp (--j * Cx w)) ==>
   z_transform (\n. ccos (Cx(w*(&n)))) z = 
   (z * z - z * ccos (Cx w)) / 
    (z * z - z * Cx (&2) * ccos (Cx w) + Cx (&1))`,
  REWRITE_TAC[j] THEN REPEAT STRIP_TAC THEN
  ABBREV_TAC
    `k = (z * z - z * ccos (Cx w)) / 
    (z * z - z * Cx (&2) * ccos (Cx w) + Cx (&1))` THEN
  REWRITE_TAC [ccos; CX_MUL] THEN
  SIMP_TAC [COMPLEX_FIELD `((a:complex) + b) / c = a / c + b / c`] THEN
  REWRITE_TAC [complex_div] THEN ONCE_REWRITE_TAC [S_LEMMA] THEN
  IMP_REWRITE_TAC [Z_TRANSFORM_LINEAR_ALT] THEN REPEAT STRIP_TAC THENL
   [SIMP_TAC
      [COMPLEX_FIELD
         `!w j.(--j * w * Cx (&n)) = 
            (--(Cx (&n) * (j * w)))`] THEN
    IMP_REWRITE_TAC [Z_TRANSFORM_NEG_EXPONETIAL; Z_TRANSFORM_POS_EXPONETIAL] THEN
    SIMP_TAC [complex_pow] THEN
    SIMP_TAC
      [COMPLEX_NORM_MUL; NORM_CEXP; RE_MUL_II; IM_CX; REAL_EXP_0;
       REAL_FIELD `-- &0 = &0`; REAL_INV_1] THEN
    REWRITE_TAC [GSYM real_div; COMPLEX_NORM_INV] THEN
    IMP_REWRITE_TAC [REAL_LT_LDIV_EQ] THEN
    ASM_SIMP_TAC [REAL_MUL_LID; REAL_FIELD `&1 < x ==> &0 < x`] THEN
    REWRITE_TAC [GSYM COMPLEX_ADD_LDISTRIB] THEN
    REWRITE_TAC [GSYM complex_div] THEN
    SIMP_TAC [COMPLEX_INV_MUL; GSYM complex_div] THEN
    SIMP_TAC [GSYM CEXP_NEG] THEN
    IMP_REWRITE_TAC
      [COMPLEX_FIELD `~(c = Cx(&0))  ==> a - b / c = (a*c - b) / c`] THEN
    SIMP_TAC [GSYM COMPLEX_NORM_NZ] THEN
    ASM_SIMP_TAC [REAL_MUL_LID; REAL_FIELD `&1 < x ==> &0 < x`] THEN
    SIMP_TAC [COMPLEX_MUL_LID] THEN
    SIMP_TAC [complex_div; COMPLEX_INV_MUL; COMPLEX_INV_INV] THEN
    SIMP_TAC [COMPLEX_MUL_LID] THEN
    SIMP_TAC [COMPLEX_FIELD `!(a:complex) (b:complex). inv a * b = b / a`] THEN
    IMP_REWRITE_TAC
      [COMPLEX_FIELD
         `~((b - c)= Cx(&0)) /\ ~((x - y)= Cx(&0)) ==>  a / (b - c) + d / (x - y) = (a*(x - y) + d*(b-c))/((b-c)*(x-y))`] THEN
    SIMP_TAC [COMPLEX_SUB_LDISTRIB] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `!j. (z * z - z * cexp (--(j * Cx w)) + z * z - z * cexp (j * Cx w)) = 
 Cx(&2)*z*z - z*(cexp (--(j * Cx w)) +  cexp (j * Cx w))`] THEN
    SIMP_TAC [COMPLEX_SUB_RDISTRIB] THEN SIMP_TAC [GSYM CEXP_ADD] THEN
    SIMP_TAC [COMPLEX_FIELD `(a:complex) + --a = Cx(&0)`; CEXP_0] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `!j. (z * z - cexp (j * Cx w) * z - (z * cexp (--(j * Cx w)) - Cx (&1))) = 
((z:complex) * z - z*(cexp (j * Cx w)  + cexp (--(j * Cx w))) + Cx (&1))`] THEN
    SIMP_TAC [complex_div] THEN
    SIMP_TAC [COMPLEX_FIELD `(a*b)*c = (a*c)*(b:complex)`] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `!j. ((Cx (&2) * z * z - z * (cexp (--(j * Cx w)) + cexp (j * Cx w))) *
  inv (Cx (&2))) = ((Cx (&2) * z * z - z * (cexp (--(j * Cx w)) + cexp (j * Cx w))) /(Cx (&2)))`] THEN
    SIMP_TAC
      [COMPLEX_FIELD
         `!j. (Cx (&2) * z * z - z * (cexp (--(j * Cx w)) + cexp (j * Cx w))) / Cx (&2)
 = 
  z * z - z * ((cexp ((j * Cx w)) + cexp (--(j * Cx w))) / Cx (&2))`] THEN
    SIMP_TAC [COMPLEX_FIELD `!j. (--(j * Cx w)) = (--j * Cx w)`] THEN
    REWRITE_TAC [GSYM ccos] THEN REWRITE_TAC [GSYM CCOS_ALT] THEN
    POP_ASSUM MP_TAC THEN SIMP_TAC [complex_div] THEN
    ASM_SIMP_TAC [COMPLEX_FIELD `~((a - b) = Cx(&0)) = ~(a = b)`];
    SIMP_TAC [roc; IN_ELIM_THM; GSYM COMPLEX_NORM_NZ] THEN
    ASM_SIMP_TAC [REAL_FIELD `&1 < x ==> &0 < x`] THEN REWRITE_TAC [summable] THEN
    SIMP_TAC [complex_div; GSYM COMPLEX_POW_INV; GSYM COMPLEX_POW_MUL] THEN
    SIMP_TAC [COMPLEX_FIELD `!w j. (j * w * Cx (&n)) = (Cx (&n) * (j * w))`] THEN
    SIMP_TAC [CEXP_N] THEN SIMP_TAC [GSYM COMPLEX_POW_MUL] THEN
    EXISTS_TAC
      `(cexp (j * Cx w) * inv z) pow 0 / (Cx (&1) - (cexp (j * Cx w) * inv z))` THEN
    REWRITE_TAC[j] THEN MATCH_MP_TAC SUMS_GP THEN
    SIMP_TAC
      [COMPLEX_NORM_MUL; NORM_CEXP; RE_MUL_II; IM_CX; REAL_EXP_0;
       REAL_FIELD `-- &0 = &0`; REAL_INV_1] THEN
    REWRITE_TAC [GSYM real_div; COMPLEX_NORM_INV];
    SIMP_TAC [roc; IN_ELIM_THM; GSYM COMPLEX_NORM_NZ] THEN
    ASM_SIMP_TAC [REAL_FIELD `&1 < x ==> &0 < x`] THEN REWRITE_TAC [summable] THEN
    SIMP_TAC [complex_div; GSYM COMPLEX_POW_INV; GSYM COMPLEX_POW_MUL] THEN
    SIMP_TAC
      [COMPLEX_FIELD `!w j. (--j * w * Cx (&n)) = (Cx (&n) * (--j * w))`] THEN
    SIMP_TAC [CEXP_N] THEN SIMP_TAC [GSYM COMPLEX_POW_MUL] THEN
    EXISTS_TAC
      `(cexp (--j * Cx w) * inv z) pow 0 / (Cx (&1) - (cexp (--j * Cx w) * inv z))` THEN
    REWRITE_TAC[j] THEN MATCH_MP_TAC SUMS_GP THEN
    SIMP_TAC
      [COMPLEX_NORM_MUL; NORM_CEXP; RE_MUL_II; IM_CX; REAL_EXP_0;
       REAL_FIELD `-- &0 = &0`; REAL_INV_1;
       COMPLEX_FIELD `!j. (--j * Cx w) = (j * --Cx (w))`; GSYM CX_NEG] THEN
    REWRITE_TAC [GSYM real_div; COMPLEX_NORM_INV; RE_CX]] THEN
  IMP_REWRITE_TAC [REAL_LT_LDIV_EQ] THEN
  ASM_SIMP_TAC [REAL_MUL_LID; REAL_FIELD `&1 < x ==> &0 < x`]);;

(*====================================================================*)


let Z_SIMP_TAC L = 
ASM_REWRITE_TAC L THEN 
(MATCH_MP_TAC Z_TRANSFORM_TIME_ADVANCE  ORELSE 
MATCH_MP_TAC ZTRANSFORM_TIME_DELAY  ORELSE 
MATCH_MP_TAC Z_TRANSFORM_ADD  ORELSE 
MATCH_MP_TAC Z_TRANSFORM_SUB  ORELSE 
MATCH_MP_TAC Z_TRANSFORM_LINEAR_ALT  ORELSE 
MATCH_MP_TAC Z_TRANSFORM_SCALING_IN_Z  ORELSE 
MATCH_MP_TAC Z_TRANSFORM_FIRST_DIFFERENCE  ORELSE 
MATCH_MP_TAC Z_TRANSFORM_SCALING_IN_Z_NEG  ORELSE 
MATCH_MP_TAC Z_TRANSFORM_CONST_MUL);;


(*===================================================================*)
(*                   DIFFERENCE EQUATION                             *)
(*===================================================================*)

let differen_eq = new_definition
    `differen_eq lst (f:num->complex) (n:num) (x:num) =
    		( vsum (0..n) (\t. (EL (t) lst) * f (x - t))) `;;

let nth_differen = new_definition
    `nth_differen lst (f:num->complex) (n:num) (x:num) =
    		( vsum (0..n) (\t. (EL (t) lst) * f (x - t))) `;;


let DIFFEN_EQ_NTH_DIFF = prove (`
!f n x.  differen_eq lst f n x =  nth_differen lst f n x `,

SIMP_TAC[differen_eq;nth_differen ]);;


let VSUM_00 = prove
 (`vsum(0..0) f = f(0)`,
  REWRITE_TAC[VSUM_SING_NUMSEG]);;

let VSUM_11 = prove
 (`!t. vsum(0..1) t = t(0) + t(1)`,
  REWRITE_TAC[num_CONV `1`; VSUM_CLAUSES_NUMSEG] THEN
  REWRITE_TAC[VSUM_SING_NUMSEG; ARITH; REAL_ADD_ASSOC]);;

let VSUM_22 = prove
 (`!t. vsum(0..2) t = t(0) + t(1) + t(2)`,
  REWRITE_TAC[num_CONV `1`; num_CONV `2`; VSUM_CLAUSES_NUMSEG] THEN
  REWRITE_TAC[VSUM_SING_NUMSEG; ARITH; VECTOR_ADD_ASSOC]);;

let VSUM_2 = prove
 (`!t. vsum(1..2) t = t(1) + t(2)`,
  REWRITE_TAC[num_CONV `1`; num_CONV `2`; VSUM_CLAUSES_NUMSEG] THEN
  REWRITE_TAC[VSUM_SING_NUMSEG; ARITH; VECTOR_ADD_ASSOC]
  THEN VECTOR_ARITH_TAC);;


let DIFF_EQ_Z_SUMMABLE_ALT = prove
 (`!N lst f z. z_tr_summable_cpow f z /\ (is_causal_signal f) ==>

z_tr_summable_cpow (\x. differen_eq lst f N x) z`,
  INDUCT_TAC THENL
   [SIMP_TAC [differen_eq; EL] THEN SIMP_TAC [VSUM_00; EL] THEN
    REPEAT STRIP_TAC;
    ASM_SIMP_TAC [differen_eq; VSUM_CLAUSES_NUMSEG; ARITH_RULE `0 <= SUC N`] THEN
    REPEAT STRIP_TAC THEN MATCH_MP_TAC Z_SUMMABLE_LINEAR THEN
    ASM_SIMP_TAC [GSYM differen_eq]] THEN
  MATCH_MP_TAC Z_SUMMABLE_CMUL_DIFF THEN ASM_MESON_TAC [is_causal_signal]);;




let DIFF_EQ_SUMMABLE = prove (`!N lst f z.
         z IN ROC f /\ is_causal_signal f
         ==> z_tr_summable_cpow (\x. differen_eq lst f N x) z`,

SIMP_TAC[GSYM ROC_Z_SUMMABLE_CPOW;DIFF_EQ_Z_SUMMABLE_ALT ]);;


let DIFF_EQ_SOLUTION_ALT = prove
 (`!f lst N z. ~(z = Cx (&0))  /\  z_tr_summable_cpow f z 
/\  (is_causal_signal f)
 ==> 
   z_tr (\x. differen_eq lst f  N x) z  =  
        z_tr f z * vsum (0..N) (\n. z cpow --(Cx(&n)) * EL n lst)`,
  GEN_TAC THEN GEN_TAC THEN INDUCT_TAC THENL
   [SIMP_TAC
      [VSUM_00; differen_eq; EL; SUB; cpow; COMPLEX_MUL_RZERO; CEXP_NEG] THEN
    SIMP_TAC
      [COMPLEX_MUL_LZERO; CEXP_NEG; CEXP_0;
       COMPLEX_FIELD `--Cx (&0) * clog z = --(Cx (&0) * clog z)`;
       COMPLEX_FIELD `inv (Cx (&1)) = Cx (&1)`; COMPLEX_MUL_RID;
       COMPLEX_MUL_RID] THEN
    REPEAT STRIP_TAC THEN IMP_REWRITE_TAC [ZTR_CONST_MUL] THEN
    SIMP_TAC [COMPLEX_MUL_LID; ETA_AX];
    REPEAT STRIP_TAC THEN
    ASM_SIMP_TAC [differen_eq; VSUM_CLAUSES_NUMSEG; ARITH_RULE `0 <= SUC N`] THEN
    REWRITE_TAC [GSYM differen_eq] THEN SIMP_TAC [COMPLEX_ADD_LDISTRIB] THEN
    ONCE_SIMP_TAC
      [COMPLEX_FIELD
         `(\x. differen_eq lst f N x + EL (SUC N) lst * f (x - SUC N)) = 
 (\x. (\x. differen_eq lst f N x) x  + (\x. EL (SUC N) lst * f (x - SUC N))x)`] THEN
    IMP_REWRITE_TAC
      [ZTR_ADD; ROC_Z_SUMMABLE_CPOW; DIFF_EQ_SUMMABLE; Z_SUMMABLE_CMUL_DIFF] THEN
    SIMP_TAC [COMPLEX_FIELD `! a b c. a + b = a + c <=> b = c`] THEN
    ONCE_SIMP_TAC
      [COMPLEX_FIELD
         `(\x. EL (SUC N) lst * f (x - SUC N)) = 
                          (\x. EL (SUC N) lst * (\x. f (x - SUC N)) x )`] THEN
    IMP_REWRITE_TAC [ZTR_CONST_MUL] THEN
    ONCE_REWRITE_TAC
      [COMPLEX_FIELD
         `!x z N f. f (x - SUC N) = 
            Cx(&1)*f (x - SUC N)`] THEN
    IMP_REWRITE_TAC [Z_SUMMABLE_CMUL_DIFF] THEN SIMP_TAC [COMPLEX_MUL_LID] THEN
    REPEAT (POP_ASSUM MP_TAC) THEN REWRITE_TAC [is_causal_signal] THEN
    REPEAT STRIP_TAC THEN ONCE_ASM_REWRITE_TAC [] THEN
    IMP_REWRITE_TAC [RIGHT_SHIFT_LEMMA]] THEN
  CONV_TAC COMPLEX_FIELD);;


let DIFF_EQ_SOLUTION_ZTR = prove (`!f lst N z.
         ~(z = Cx (&0)) /\ z IN ROC f /\ (is_causal_signal f)
         ==> z_tr (\x. differen_eq lst f N x) z =
             z_tr f z * vsum (0..N) (\n. z cpow --Cx (&n) * EL n lst)`,

SIMP_TAC[GSYM ROC_Z_SUMMABLE_CPOW;DIFF_EQ_SOLUTION_ALT]);;


let DIFF_EQ_SOLUTION = prove(
        `!f lst N z.
             z IN roc f /\ (is_causal_signal f)
         ==> z_transform (\x. differen_eq lst f N x) z =
             z_transform f z * vsum (0..N) (\n. z cpow --Cx (&n) * EL n lst)`,


ZTR_ZTRANSFORM_TAC [DIFF_EQ_SOLUTION_ZTR]);;



let DIFF_EQ_SOLUTION_NTH_DIFF = prove(
        `!f lst N z.
             z IN roc f /\ (is_causal_signal f)
         ==> z_transform (\x. nth_differen lst f N x) z =
             z_transform f z * vsum (0..N) (\n. z cpow --Cx (&n) * EL n lst)`,


SIMP_TAC [GSYM DIFFEN_EQ_NTH_DIFF;DIFF_EQ_SOLUTION]);;


let  DIFF_EQ_SOLUTION_ABST = 
     SIMP_RULE [ETA_AX;FUN_EQ_THM]  DIFF_EQ_SOLUTION;;


(*======================================================================*)
(*                    Transfer Function                                 *)
(*======================================================================*)


let VSUM_COMBINE_FROM_0 = prove(
`! m f. vsum (0..m) f = vsum (0..0) f + vsum (0 + 1..m) f`,

REPEAT GEN_TAC THEN 
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN 
MATCH_MP_TAC VSUM_COMBINE_R THEN 
CONV_TAC ARITH_RULE);;

let VSUM_O_FUNC = prove (` !s. FINITE s ==> 
    vsum s (\n. (Cx(&0))) =  Cx (&0)`,

ONCE_REWRITE_TAC[COMPLEX_FIELD `!n. Cx(&0) = Cx(&0)* Cx(&n)`] THEN 
SIMP_TAC[VSUM_COMPLEX_LMUL] THEN 
SIMP_TAC[COMPLEX_MUL_LZERO]);;

let CPOW_POW = prove (
   ` !n w. ~(w = Cx(&0)) ==> w cpow Cx(&n) = w pow n`,

INDUCT_TAC THENL[
GEN_TAC THEN 
SIMP_TAC[cpow;complex_pow;COMPLEX_MUL_LZERO;CEXP_0];

SIMP_TAC[cpow;complex_pow;COMPLEX_MUL_LZERO;CEXP_0; ADD1;
       CX_ADD;GSYM REAL_OF_NUM_ADD;COMPLEX_ADD_RDISTRIB;
      COMPLEX_MUL_LID;GSYM CEXP_ADD] THEN 
REWRITE_TAC[MESON[GSYM CEXP_ADD]`cexp ((Cx (&n) * clog w) + clog w) 
            = cexp (Cx (&n) * clog w)* cexp(clog w)`] THEN
SIMP_TAC[CEXP_CLOG; 
           COMPLEX_FIELD `w * w pow n =  (w pow n) * w`] THEN 
SIMP_TAC[COMPLEX_EQ_MUL_RCANCEL] THEN 
POP_ASSUM MP_TAC THEN 
SIMP_TAC[cpow]]);;


let INV_CPOW = prove (
   `!n z. ~(z = Cx(&0)) ==>  z cpow --Cx(&n) = (inv z) cpow Cx(&n)`,  

INDUCT_TAC THENL[
REWRITE_TAC[CPOW_NEG] THEN 
SIMP_TAC[cpow; COMPLEX_FIELD ` !z. 
                     ~(z = Cx (&0)) ==> ~(inv z = Cx (&0))`] THEN 
SIMP_TAC[COMPLEX_MUL_LZERO;CEXP_0] THEN 
CONV_TAC COMPLEX_FIELD;

REWRITE_TAC[ADD1;GSYM REAL_OF_NUM_ADD;CX_ADD] THEN 
SIMP_TAC[COMPLEX_FIELD 
         `--(Cx (&n) + Cx (&1))= (--Cx (&n) + --Cx (&1)) `;
         GSYM CX_NEG; CPOW_ADD] THEN 
REWRITE_TAC[CX_NEG] THEN 
ASM_SIMP_TAC[] THEN 
SIMP_TAC[COMPLEX_EQ_MUL_LCANCEL] THEN 
REPEAT STRIP_TAC THEN 
DISJ2_TAC THEN
ASM_SIMP_TAC[cpow; COMPLEX_FIELD ` !z. 
                     ~(z = Cx (&0)) ==> ~(inv z = Cx (&0))`] THEN 
SIMP_TAC[COMPLEX_MUL_LID; 
        COMPLEX_FIELD `--Cx (&1) * clog z= --(Cx (&1) * clog z)`] THEN 
ASM_SIMP_TAC[CEXP_CLOG;CEXP_NEG; COMPLEX_FIELD ` !z. 
                     ~(z = Cx (&0)) ==> ~(inv z = Cx (&0))`]]);;



let VSUM_Z_NZ = prove (`
   !N z lst. FINITE (0..N) /\ (!x. EL x lst = Cx (&0))  ==>
     (vsum (0..N) (\n. z cpow --Cx (&n) * EL n lst) = Cx (&0)) `,
REPEAT STRIP_TAC THEN 
ASSERT_TAC `Cx (&0) = vsum (0..N) (\n. (Cx(&0))) ` THENL[
ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN 
MATCH_MP_TAC VSUM_O_FUNC THEN 
ASM_SIMP_TAC[];ALL_TAC] THEN  
ONCE_ASM_REWRITE_TAC[] THEN 
MATCH_MP_TAC VSUM_EQ THEN 
REPEAT STRIP_TAC THEN 
ASM_SIMP_TAC[COMPLEX_ENTIRE]);;

let DIFF_EQ_VSUM_EXPAND = prove (` !N lst. 
  vsum (0..N) (\n. z cpow --Cx (&n) * EL n lst) =
  vsum (0..0) (\n. z cpow --Cx (&n) * EL n lst) +
  vsum (1..N) (\n. z cpow --Cx (&n) * EL n lst)`,

ONCE_REWRITE_TAC[ARITH_RULE `1 = 0 + 1`] THEN 
SIMP_TAC[GSYM VSUM_COMBINE_FROM_0]);;


let VSUM_EXPAND_ALT = prove (
    ` !m f. vsum (1..m) f =   vsum (0..m) f  - vsum (0..0) f`,

SIMP_TAC[COMPLEX_FIELD `a = b - c <=> b = c + a`] THEN 
ONCE_REWRITE_TAC[ARITH_RULE `1 = 0 + 1`  ] THEN 
MESON_TAC[ VSUM_COMBINE_FROM_0 ]);;



let DIFF_EQ_Z_EXPAND = prove (` !f lst N.
   ~(z = Cx (&0)) /\z_tr_summable_cpow f z /\ (is_causal_signal f) ==>
z_tr (\x. differen_eq lst f N x) z = 
z_tr f z * ( HD lst + vsum (1..N) (\n. z cpow --Cx (&n) * EL n lst))`,

SIMP_TAC[DIFF_EQ_SOLUTION_ALT] THEN 
ONCE_REWRITE_TAC[DIFF_EQ_VSUM_EXPAND] THEN 
REWRITE_TAC[VSUM_00] THEN 
SIMP_TAC[cpow;EL;COMPLEX_FIELD `--Cx (&0) * clog z = Cx(&0)`;
           CEXP_0; COMPLEX_MUL_LID ]);;


let SML = prove 
(`!a b x z . ~(z_tr x z = Cx(&0)) ==> (a = b <=> 
      a*(z_tr x z / z_tr x z) = b*(z_tr x z / z_tr x z))`,

CONV_TAC COMPLEX_FIELD );;
 

let SML1 = prove 
(`!a b x z . ~
((HD lst + vsum (1..N) (\n. z cpow --Cx (&n) * EL n outlst)) = 
  Cx(&0)) ==> (a = b <=> 
      a*( (HD lst + vsum (1..N) (\n. z cpow --Cx (&n) * EL n outlst))/ 
(HD lst + vsum (1..N) (\n. z cpow --Cx (&n) * EL n outlst))) = 
b*( (HD lst + vsum (1..N) (\n. z cpow --Cx (&n) * EL n outlst))/ 
(HD lst + vsum (1..N) (\n. z cpow --Cx (&n) * EL n outlst))))`,

CONV_TAC COMPLEX_FIELD );;
 

let transfer_function  =  new_definition ` 
    transfer_function  x y z = z_transform y z / z_transform x z`;;


let lccde = define `
    lccde x y alist blist M N n <=> 
    y n  = (differen_eq alist y M n) + 
       	   (differen_eq (blist) x N n) /\ (HD alist = Cx (&0))`;;

let lccde_roc  = define `
    lccde_roc g f M alst = ROC f INTER ROC g DIFF 
    { z | (Cx (&1) -  vsum (1..M) (\n. EL n alst * z cpow (--Cx(&n)))) = 
    Cx(&0)}`;;


let is_causal_lccde  = define `
    is_causal_lccde x y  <=>  (is_causal_signal x) /\
                              (is_causal_signal y)`;;

let LCCDE_TRANSFER_FUNCTION_ZTR = prove
 (`!x y inlst outlst M N.
         ~(z = Cx (&0)) /\
         z IN lccde_roc x y M alst /\
         is_causal_lccde x y /\
         (!n. lccde x y alst blst M N n) /\
         ~(z_tr x z = Cx (&0))
         ==> z_tr y z / z_tr x z =
             vsum (0..N) (\n. z cpow --Cx (&n) * EL n blst) /
             (Cx (&1) - vsum (1..M) (\n. z cpow --Cx (&n) * EL n alst))`,
  REWRITE_TAC [IN; IN_ELIM_THM; INTER; DIFF; lccde_roc; ROC; is_causal_lccde] THEN
  REWRITE_TAC [GSYM z_tr_summable_cpow; lccde] THEN REPEAT STRIP_TAC THEN
  ONCE_REWRITE_TAC [COMPLEX_MUL_SYM] THEN
  ASM_SIMP_TAC
    [COMPLEX_FIELD
       `! a b c d. ~(d = Cx(&0)) /\  ~(b = Cx(&0)) ==> ( (a / b = c / d ) = (a*d = b*c ))`] THEN
  REWRITE_TAC
    [COMPLEX_FIELD
       `! n. EL n blst * z cpow --Cx (&n) = 
                           z cpow --Cx (&n) *  EL n blst`] THEN
  ASM_SIMP_TAC [GSYM DIFF_EQ_SOLUTION_ALT; ROC_Z_SUMMABLE_CPOW] THEN
  REWRITE_TAC [COMPLEX_MUL_RID; COMPLEX_SUB_LDISTRIB] THEN
  SIMP_TAC [COMPLEX_FIELD `a - b = c <=> a = c  + b`] THEN
  REWRITE_TAC [VSUM_EXPAND_ALT] THEN REWRITE_TAC [COMPLEX_SUB_LDISTRIB] THEN
  REWRITE_TAC [VSUM_00; EL] THEN ASM_SIMP_TAC [COMPLEX_MUL_RZERO; SUB_0] THEN
  SIMP_TAC [COMPLEX_FIELD `a - Cx(&0) = a`] THEN
  ASM_SIMP_TAC [GSYM DIFF_EQ_SOLUTION_ALT; ROC_Z_SUMMABLE_CPOW] THEN
  IMP_REWRITE_TAC [GSYM ZTR_ADD] THEN ASM_SIMP_TAC [DIFF_EQ_Z_SUMMABLE_ALT] THEN
  ASSERT_TAC `z_tr y z  = z_tr (\x. y x) z` THENL
   [SIMP_TAC [ETA_AX]; ALL_TAC] THEN
  ONCE_ASM_REWRITE_TAC [] THEN ONCE_ASM_REWRITE_TAC [] THEN
  SIMP_TAC [COMPLEX_ADD_SYM]);;


let LCCDE_ROC  = new_definition `
    LCCDE_ROC g f M alst = roc f INTER roc g DIFF 
    { z | (Cx (&1) -  vsum (1..M) (\n. EL n alst * z cpow (--Cx(&n)))) = 
    Cx(&0)}`;;


let LCCDE_ROC_IMPLIES = prove

(` !g f M alst. z IN LCCDE_ROC g f M alst ==>
                z IN lccde_roc g f M alst`,

REPEAT GEN_TAC THEN 
ZTR_ZTRANSFORM_TAC[LCCDE_ROC;lccde_roc;IN_INTER;IN_DIFF]);;



let LCCDE_ROC_IMPLIES_Z_NOT_ZERO  = prove

(`!g f M alst. z IN LCCDE_ROC g f M alst ==> ~(z = Cx(&0))`,
REPEAT GEN_TAC THEN
ZTR_ZTRANSFORM_TAC[LCCDE_ROC;lccde_roc;IN_INTER;IN_DIFF]);;


 
let LCCDE_TRANSFER_FUNCTION = prove
 (`!x y inlst outlst M N.
         z IN LCCDE_ROC x y M alst /\
         is_causal_lccde x y /\
         (!n. lccde x y alst blst M N n) /\
         ~(z_transform x z = Cx (&0))
         ==> z_transform y z / z_transform x z =
             vsum (0..N) (\n. z cpow --Cx (&n) * EL n blst) /
             (Cx (&1) - vsum (1..M) (\n. z cpow --Cx (&n) * EL n alst))`,
  REPEAT GEN_TAC THEN REPEAT STRIP_TAC THEN
  IMP_REWRITE_TAC [ZTR_EQ; LCCDE_ROC_IMPLIES_Z_NOT_ZERO] THEN
  IMP_REWRITE_TAC [LCCDE_TRANSFER_FUNCTION_ZTR] THEN
  ASM_SIMP_TAC [LCCDE_ROC_IMPLIES_Z_NOT_ZERO; LCCDE_ROC_IMPLIES; GSYM ZTR_EQ] THEN
  ASM_MESON_TAC [LCCDE_ROC_IMPLIES_Z_NOT_ZERO; GSYM ZTR_EQ]);;



(*=======================================================================*)
(*                   z-Transform Uniqueness                              *)
(*=======================================================================*)

let exterior_circle = new_definition `
    exterior_circle s = 
    (?R. &0 < R /\ (!(z:complex). R < dist (z,Cx(&0)) ==> z IN s))`;;


let EXTERIOR_CIRCLE = prove(`
     !f. exterior_circle s =  
    (?R. &0 < R /\ (!(z:complex). R < norm z ==> z IN s))`,

MESON_TAC[exterior_circle;dist;COMPLEX_FIELD ` !z. z - Cx(&0) = z`]);;


let EXTERIOR_CIRCLE_INTER = prove
 (`!f g. (exterior_circle (roc f) /\ exterior_circle (roc g)) ==> 
          exterior_circle  ((roc f) INTER (roc g))`,
  SIMP_TAC [exterior_circle; roc; IN_ELIM_THM; INTER] THEN REPEAT STRIP_TAC THEN
  EXISTS_TAC `max R R'` THEN REPEAT STRIP_TAC THENL
   [REWRITE_TAC [real_max] THEN COND_CASES_TAC THEN
    IMP_REWRITE_TAC [REAL_LTE_TRANS];
    ASSERT_TAC `R' < dist (z,Cx (&0)) /\ R < dist (z,Cx (&0))` THENL
     [IMP_REWRITE_TAC [REAL_LET_TRANS; REAL_MAX_MAX]; ASM_MESON_TAC []];
    ASSERT_TAC `R' < dist (z,Cx (&0)) /\ R < dist (z,Cx (&0))` THENL
     [IMP_REWRITE_TAC [REAL_LET_TRANS; REAL_MAX_MAX]; ASM_MESON_TAC []];
    ASSERT_TAC `R' < dist (z,Cx (&0)) /\ R < dist (z,Cx (&0))` THENL
     [IMP_REWRITE_TAC [REAL_LET_TRANS; REAL_MAX_MAX]; ASM_MESON_TAC []];
    ASSERT_TAC `R' < dist (z,Cx (&0)) /\ R < dist (z,Cx (&0))` THENL
     [IMP_REWRITE_TAC [REAL_LET_TRANS; REAL_MAX_MAX]; ASM_MESON_TAC []]]);;


let AB_LEMMA = prove 
(`!M s a.
           summable s a /\ &0 < M
           ==> (!z. norm z < &1 ==> summable s (\i. a i * z pow i))`,
MESON_TAC[
  REWRITE_RULE [COMPLEX_POW_ONE;COMPLEX_MUL_RID;COMPLEX_NORM_CX;REAL_ABS_1;ETA_AX]
(ISPECL [`M:real`;`s:num->bool`;`a:num->real^2`;`Cx(&1)`] ABEL_LIMIT_THEOREM)] );;
  

let EXTERIOR_ROC_SUMMABLE_INPUT = prove
 (`!f. summable (from 0) f ==>
       exterior_circle (roc f)`,
  GEN_TAC THEN SIMP_TAC [roc; EXTERIOR_CIRCLE; IN_ELIM_THM; Z_TRANSFORM] THEN
  DISCH_TAC THEN SIMP_TAC [complex_div; GSYM COMPLEX_POW_INV] THEN
  EXISTS_TAC `&1` THEN REPEAT STRIP_TAC THENL
   [ALL_TAC;
    UNDISCH_TAC `&1 < norm (z:complex)` THEN ASM_SIMP_TAC [] THEN
    SIMP_TAC [COMPLEX_NORM_0];
    MP_TAC (ISPECL [`&1`; `(from 0):num->bool`; `f:num->complex`] AB_LEMMA) THEN
    ANTS_TAC THENL
     [ASM_SIMP_TAC [REAL_FIELD `&0 < &1`];
      DISCH_THEN (MP_TAC o SPEC `(inv z):complex`) THEN
      ASM_MESON_TAC [REAL_INV_LT_1; COMPLEX_NORM_INV]]] THEN
  REAL_ARITH_TAC);;


let EXTERIOR_ROC_SUMMABLE_RADIUS_OF_CONVERGENCE = prove
 
(`!f c N.
         c < &1 /\ (!n. n >= N ==> norm (f (SUC n)) <= c * norm (f n))
      ==> exterior_circle (roc f)`,
  REPEAT GEN_TAC THEN REPEAT STRIP_TAC THEN
  MATCH_MP_TAC EXTERIOR_ROC_SUMMABLE_INPUT THEN REWRITE_TAC [summable] THEN
  IMP_REWRITE_TAC [SERIES_RATIO]);;


let HIGHER_COMPLEX_DERIVATIVE_POWER_SERIES_GEN = prove 

(`!f c r n k.
         &0 < r /\
         n IN k /\
         (!w. dist (w,z) < r ==> ((\i. c i * (w - z) pow i) sums f w) k)
         ==> higher_complex_derivative n f z / Cx (&(FACT n)) = c n`,

REWRITE_TAC[HIGHER_COMPLEX_DERIVATIVE_POWER_SERIES]);;


let Z_TRANSFORM_UNIQUE = prove
 (`!f g.  exterior_circle (roc f) /\  exterior_circle (roc g) ==>
         (!z. z_transform f z = z_transform g z) = (!n. f n  = g n)`,
  REWRITE_TAC [EXTERIOR_CIRCLE; roc; IN_ELIM_THM] THEN REPEAT STRIP_TAC THEN
  EQ_TAC THENL [ALL_TAC; SIMP_TAC [ETA_AX; z_transform]] THEN
  REPEAT STRIP_TAC THEN
  ASSERT_TAC
    `!n.
    higher_complex_derivative n (\z. z_transform f  (inv z)) (Cx(&0)) / Cx (&(FACT n)) = f n` THENL
   [REPEAT GEN_TAC THEN
    MATCH_MP_TAC HIGHER_COMPLEX_DERIVATIVE_POWER_SERIES_GEN THEN
    EXISTS_TAC `&1/R` THEN EXISTS_TAC `(from 0)` THEN UNDISCH_TAC `&0 < R` THEN
    SIMP_TAC
      [IN_FROM; ARITH_RULE `0 <= n`; REAL_LT_DIV; REAL_FIELD `&0 < &1`; dist;
       COMPLEX_FIELD `!(a:complex). a - Cx(&0) = a`; z_transform;
       complex_div; COMPLEX_INV_INV; GSYM COMPLEX_POW_INV; SUMS_INFSUM] THEN
    REPEAT STRIP_TAC THEN
    UNDISCH_TAC
      `!z. R < norm z
          ==> ~(z = Cx (&0)) /\ summable (from 0) (\n. f n / z pow n)` THEN
    DISCH_TAC THEN POP_ASSUM (MP_TAC o SPEC `inv (w:complex)`) THEN DISCH_TAC THEN
    ASM_CASES_TAC `w = Cx(&0)` THENL
     [ASM_SIMP_TAC [] THEN MATCH_MP_TAC SUMMABLE_FROM_ELSEWHERE THEN
      EXISTS_TAC `(n:num) + 1` THEN REWRITE_TAC [GSYM SUMMABLE_REINDEX] THEN
      SIMP_TAC [ARITH_RULE `~((x:num) + 1 = 0)`; COMPLEX_POW_ZERO] THEN
      SIMP_TAC [COMPLEX_MUL_RZERO; GSYM COMPLEX_VEC_0; SUMMABLE_0];
      POP_ASSUM MP_TAC THEN REWRITE_TAC [GSYM COMPLEX_NORM_NZ] THEN STRIP_TAC THEN
      ASSERT_TAC `R < norm (inv w)` THENL
       [ONCE_REWRITE_TAC [REAL_FIELD `!(a:real) b. a < b <=> b > a`] THEN
        SIMP_TAC [REAL_FIELD `!(x:real). x > e <=>  e < x`] THEN
        ONCE_REWRITE_TAC [COMPLEX_NORM_INV] THEN
        ONCE_REWRITE_TAC [REAL_FIELD `!w. inv (norm w) = &1* inv (norm w)`] THEN
        SIMP_TAC [GSYM real_div] THEN IMP_REWRITE_TAC [REAL_LT_RDIV_EQ] THEN
        ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
        ASM_SIMP_TAC [GSYM REAL_LT_RDIV_EQ];
        ASM_SIMP_TAC [GSYM REAL_LT_RDIV_EQ] THEN
        UNDISCH_TAC
          `R < norm (inv w)
      ==> ~(inv w = Cx (&0)) /\ summable (from 0) (\n. f n / inv w pow n)` THEN
        ASM_SIMP_TAC [complex_div; COMPLEX_INV_INV; GSYM COMPLEX_POW_INV]]];
    ASSERT_TAC
      `!n.
    higher_complex_derivative n (\z. z_transform g  (inv z)) (Cx(&0)) / Cx (&(FACT n)) = g n` THENL
     [ALL_TAC; REPEAT (POP_ASSUM MP_TAC) THEN SIMP_TAC []] THEN
    REPEAT GEN_TAC THEN
    MATCH_MP_TAC HIGHER_COMPLEX_DERIVATIVE_POWER_SERIES_GEN THEN
    EXISTS_TAC `&1/R'` THEN EXISTS_TAC `(from 0)` THEN UNDISCH_TAC `&0 < R'` THEN
    SIMP_TAC
      [IN_FROM; ARITH_RULE `0 <= n`; REAL_LT_DIV; REAL_FIELD `&0 < &1`; dist;
       COMPLEX_FIELD `!(a:complex). a - Cx(&0) = a`; z_transform;
       complex_div; COMPLEX_INV_INV; GSYM COMPLEX_POW_INV; SUMS_INFSUM] THEN
    REPEAT STRIP_TAC THEN
    UNDISCH_TAC
      `!z. R' < norm z
          ==> ~(z = Cx (&0)) /\ summable (from 0) (\n. g n / z pow n)` THEN
    DISCH_TAC THEN POP_ASSUM (MP_TAC o SPEC `inv (w:complex)`) THEN DISCH_TAC THEN
    ASM_CASES_TAC `w = Cx(&0)` THENL
     [ASM_SIMP_TAC [] THEN MATCH_MP_TAC SUMMABLE_FROM_ELSEWHERE THEN
      EXISTS_TAC `(n:num) + 1` THEN REWRITE_TAC [GSYM SUMMABLE_REINDEX] THEN
      SIMP_TAC [ARITH_RULE `~((x:num) + 1 = 0)`; COMPLEX_POW_ZERO] THEN
      SIMP_TAC [COMPLEX_MUL_RZERO; GSYM COMPLEX_VEC_0; SUMMABLE_0];
      POP_ASSUM MP_TAC THEN REWRITE_TAC [GSYM COMPLEX_NORM_NZ] THEN STRIP_TAC THEN
      ASSERT_TAC `R' < norm (inv w)` THENL
       [ONCE_REWRITE_TAC [REAL_FIELD `!(a:real) b. a < b <=> b > a`] THEN
        SIMP_TAC [REAL_FIELD `!(x:real). x > e <=>  e < x`] THEN
        ONCE_REWRITE_TAC [COMPLEX_NORM_INV] THEN
        ONCE_REWRITE_TAC [REAL_FIELD `!w. inv (norm w) = &1* inv (norm w)`] THEN
        SIMP_TAC [GSYM real_div] THEN IMP_REWRITE_TAC [REAL_LT_RDIV_EQ] THEN
        ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
        ASM_SIMP_TAC [GSYM REAL_LT_RDIV_EQ];
        ASM_SIMP_TAC [GSYM REAL_LT_RDIV_EQ] THEN
        UNDISCH_TAC
          `R' < norm (inv w)
      ==> ~(inv w = Cx (&0)) /\ summable (from 0) (\n. g n / inv w pow n)` THEN
        ASM_SIMP_TAC [complex_div; COMPLEX_INV_INV; GSYM COMPLEX_POW_INV]]]]);;



let Z_TRANSFORM_UNIQUE_ABST = 
SIMP_RULE [ETA_AX;GSYM FUN_EQ_THM]  Z_TRANSFORM_UNIQUE;;


let Z_IMP_EQ = 
REWRITE_RULE [
MESON []`
(A ==> (a = b) = (c = d)) <=>
((A /\ (a = b) ==> (c = d)) /\ ( A /\ (c = d) ==> (a = b)))`]
 Z_TRANSFORM_UNIQUE_ABST;;


let Z_UNIQ_SND = 
MESON [ Z_IMP_EQ]` 
!f g. (exterior_circle (roc f) /\ exterior_circle (roc g)) /\ f = g
          ==> z_transform f = z_transform g`;;

let Z_UNIQ_ETA =
SIMP_RULE [ETA_AX;GSYM FUN_EQ_THM]  Z_TRANSFORM_UNIQUE;;


(*======================================================*)
(*                  CONVOLUTION                         *)
(*======================================================*)
let convl = new_definition `
    convl f g = (\n. infsum (from 0) (\l. (f l) * g (n-l)))`;; 

let INFSUM_FINITE = prove
 (`!f s. FINITE s ==> infsum s f = vsum s f`,
REPEAT STRIP_TAC THEN MATCH_MP_TAC INFSUM_UNIQUE THEN ASM_SIMP_TAC[SERIES_FINITE]);;

let INFSUM_FINITE_SUPPORT = prove
 (`!f:num->real^N s k.
     FINITE (s INTER k) /\ (!x. x IN k /\ ~(x IN s) ==> f x = vec 0)
     ==> infsum k f = vsum (s INTER k) f`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC INFSUM_UNIQUE THEN ASM_SIMP_TAC[SERIES_FINITE_SUPPORT]);;

let  CONVL_CAUSAL = prove
 (`!f g. is_causal_signal g ==> convl f g n = vsum (0..n) (\l. f l * g (n-l))`,

REWRITE_TAC [is_causal_signal;convl] THEN REPEAT STRIP_TAC THEN
ONCE_ASM_REWRITE_TAC[] THEN IMP_REWRITE_TAC[INFSUM_FINITE_SUPPORT] THEN EXISTS_TAC `0..n` THEN
REWRITE_TAC[ONCE_REWRITE_RULE[INTER_COMM]FROM_INTER_NUMSEG;FINITE_NUMSEG;COMPLEX_VEC_0]
THEN SIMP_TAC [numseg;IN_ELIM_THM;ARITH_RULE `~(0 <= x /\ x <= n) <=> ~(x <= n)`;COMPLEX_MUL_RZERO]);;

(*======================================================*)
(*                Initial Value Theorem                 *)
(*======================================================*)

let  INFSUM_0_1 = prove

(`!f. summable (from 0) f ==>
infsum (from 0) f = vsum (0..1 - 1) f + infsum (from 1) f`,
REPEAT STRIP_TAC  THEN MATCH_MP_TAC INFSUM_OFFSET THEN 
ASM_SIMP_TAC[ARITH_RULE `(0:num) < 1 /\ (0:num) <= 1` ]);;


let SUMMABLE_FUN_MUL_RZERO_POW = prove 
(` !f k. summable (from k) (\n. f n * Cx (&0) pow n)`,

REPEAT STRIP_TAC THEN 
MATCH_MP_TAC SUMMABLE_FROM_ELSEWHERE THEN 
EXISTS_TAC `(k:num) + 1 ` THEN 
REWRITE_TAC[GSYM SUMMABLE_REINDEX] THEN 
SIMP_TAC[ARITH_RULE `~((k:num) + 1 = 0)`;COMPLEX_POW_ZERO] THEN
SIMP_TAC[COMPLEX_MUL_RZERO;GSYM COMPLEX_VEC_0;SUMMABLE_0]);;

let INFSUM_FUN_MUL_RZERO_POW = prove 
(`!f k. infsum (from 1) (\n. f n * Cx (&0) pow n) = Cx (&0)`,

REPEAT STRIP_TAC THEN MATCH_MP_TAC INFSUM_UNIQUE THEN
ONCE_REWRITE_TAC[ARITH_RULE ` 1 = 0 + (1:num)`] THEN
SIMP_TAC[GSYM SUMS_REINDEX] THEN
SIMP_TAC[ARITH_RULE `~((x:num) + 1 = 0)`;COMPLEX_POW_ZERO] THEN
SIMP_TAC[COMPLEX_MUL_RZERO;GSYM COMPLEX_VEC_0;SERIES_0]);;


let Z_INITIAL_VALUE_THEOREM = prove
 (`!f. exterior_circle (roc f)  ==>
     	 ((\z. z_transform f z) --> f 0) at_infinity`,
  REWRITE_TAC
    [EXTERIOR_CIRCLE; roc; IN_ELIM_THM; IN_INTER; interior; SUBSET; open_def] THEN
  REPEAT STRIP_TAC THEN MATCH_MP_TAC LIM_ZERO_INFINITY_COMPLEX THEN
  SIMP_TAC
    [z_transform; complex_div; COMPLEX_INV_INV; GSYM COMPLEX_POW_INV;
     COMPLEX_MUL_LID] THEN
  MATCH_MP_TAC LIM_CONTINUOUS THEN SIMP_TAC [NETLIMIT_AT] THEN
  IMP_REWRITE_TAC [INFSUM_0_1] THEN
  SIMP_TAC
    [SUMMABLE_FUN_MUL_RZERO_POW; ARITH_RULE `1 - 1 = (0:num)`; VSUM_00;
     complex_pow; COMPLEX_MUL_RID;
     COMPLEX_FIELD `!a b. (a:complex) + b = a <=> b = Cx(&0)`;
     INFSUM_FUN_MUL_RZERO_POW] THEN
  MATCH_MP_TAC COMPLEX_DIFFERENTIABLE_IMP_CONTINUOUS_AT THEN
  MATCH_MP_TAC HOLOMORPHIC_ON_IMP_DIFFERENTIABLE_AT THEN
  EXISTS_TAC `ball(Cx(&0), &1/R)` THEN SIMP_TAC [OPEN_BALL] THEN
  REPEAT STRIP_TAC THENL
   [ALL_TAC;
    ASM_SIMP_TAC
      [ball; dist; IN_ELIM_THM; COMPLEX_FIELD `Cx (&0) - Cx (&0) = Cx(&0)`;
       GSYM COMPLEX_VEC_0; NORM_0; VECTOR_ARITH `vec 0 - vec 0 = vec 0`] THEN
    POP_AND_SIMP [REAL_LT_DIV; REAL_FIELD `&0 < &1`]] THEN
  SIMP_TAC [HOLOMORPHIC_IFF_POWER_SERIES] THEN REPEAT STRIP_TAC THEN
  ASSERT_TAC
    `!n. higher_complex_derivative n (\x. infsum (from 0) (\n. f n * x pow n))
       (Cx (&0)) /
       Cx (&(FACT n)) = f n` THENL
   [REPEAT GEN_TAC THEN
    MATCH_MP_TAC HIGHER_COMPLEX_DERIVATIVE_POWER_SERIES_GEN THEN
    EXISTS_TAC `&1/R` THEN EXISTS_TAC `(from 0)` THEN UNDISCH_TAC `&0 < R` THEN
    SIMP_TAC
      [IN_FROM; ARITH_RULE `0 <= n`; REAL_LT_DIV; REAL_FIELD `&0 < &1`; dist;
       COMPLEX_FIELD `!(a:complex). a - Cx(&0) = a`; SUMS_INFSUM] THEN
    REPEAT STRIP_TAC THEN
    UNDISCH_TAC
      `!z. R < norm z ==> ~(z = Cx (&0)) /\summable (from 0) (\n. f n / z pow n)` THEN
    DISCH_TAC THEN POP_ASSUM (MP_TAC o SPEC `inv (w':complex)`) THEN
    DISCH_TAC THEN ASM_CASES_TAC `w' = Cx(&0)` THENL
     [ASM_SIMP_TAC [] THEN MATCH_MP_TAC SUMMABLE_FROM_ELSEWHERE THEN
      EXISTS_TAC `(n:num) + 1` THEN REWRITE_TAC [GSYM SUMMABLE_REINDEX] THEN
      SIMP_TAC [ARITH_RULE `~((x:num) + 1 = 0)`; COMPLEX_POW_ZERO] THEN
      SIMP_TAC [COMPLEX_MUL_RZERO; GSYM COMPLEX_VEC_0; SUMMABLE_0];
      POP_ASSUM MP_TAC THEN REWRITE_TAC [GSYM COMPLEX_NORM_NZ] THEN STRIP_TAC THEN
      ASSERT_TAC `R < norm (inv w')` THENL
       [ALL_TAC;
        UNDISCH_TAC
          `R < norm (inv w')
      ==> ~(inv w' = Cx (&0)) /\ summable (from 0) (\n. f n / inv w' pow n)` THEN
        ASM_SIMP_TAC [complex_div; COMPLEX_INV_INV; GSYM COMPLEX_POW_INV]] THEN
      ONCE_REWRITE_TAC [REAL_FIELD `!(a:real) b. a < b <=> b > a`] THEN
      SIMP_TAC [REAL_FIELD `!(x:real). x > e <=>  e < x`] THEN
      ONCE_REWRITE_TAC [COMPLEX_NORM_INV] THEN
      ONCE_REWRITE_TAC [REAL_FIELD `!w'. inv (norm w') = &1* inv (norm w')`] THEN
      SIMP_TAC [GSYM real_div] THEN IMP_REWRITE_TAC [REAL_LT_RDIV_EQ] THEN
      ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
      ASM_SIMP_TAC [GSYM REAL_LT_RDIV_EQ]];
    ASM_REWRITE_TAC [] THEN
    SIMP_TAC
      [REAL_FIELD `&0 < &1`; dist;
       COMPLEX_FIELD `!(a:complex). a - Cx(&0) = a`; SUMS_INFSUM] THEN
    POP_AND_SIMP
      [ball; dist; IN_ELIM_THM; NORM_SUB; COMPLEX_FIELD `!w. w - Cx(&0) = w`] THEN
    UNDISCH_TAC
      `!z. R < norm z ==> ~(z = Cx (&0)) /\summable (from 0) (\n. f n / z pow n)` THEN
    DISCH_TAC THEN POP_ASSUM (MP_TAC o SPEC `inv (w:complex)`) THEN DISCH_TAC THEN
    ASM_CASES_TAC `w = Cx(&0)` THENL
     [ASM_SIMP_TAC [] THEN MATCH_MP_TAC SUMMABLE_FROM_ELSEWHERE THEN
      EXISTS_TAC `(n:num) + 1` THEN REWRITE_TAC [GSYM SUMMABLE_REINDEX] THEN
      SIMP_TAC [ARITH_RULE `~((x:num) + 1 = 0)`; COMPLEX_POW_ZERO] THEN
      SIMP_TAC [COMPLEX_MUL_RZERO; GSYM COMPLEX_VEC_0; SUMMABLE_0];
      POP_ASSUM MP_TAC THEN REWRITE_TAC [GSYM COMPLEX_NORM_NZ] THEN STRIP_TAC THEN
      ASSERT_TAC `R < norm (inv w)` THENL
       [ALL_TAC;
        UNDISCH_TAC
          `R < norm (inv w)
      ==> ~(inv w = Cx (&0)) /\ summable (from 0) (\n. f n / inv w pow n)` THEN
        ASM_SIMP_TAC [complex_div; COMPLEX_INV_INV; GSYM COMPLEX_POW_INV]] THEN
      ONCE_REWRITE_TAC [REAL_FIELD `!(a:real) b. a < b <=> b > a`] THEN
      SIMP_TAC [REAL_FIELD `!(x:real). x > e <=>  e < x`] THEN
      ONCE_REWRITE_TAC [COMPLEX_NORM_INV] THEN
      ONCE_REWRITE_TAC [REAL_FIELD `!w. inv (norm w) = &1* inv (norm w)`] THEN
      SIMP_TAC [GSYM real_div] THEN IMP_REWRITE_TAC [REAL_LT_RDIV_EQ] THEN
      ONCE_REWRITE_TAC [REAL_MUL_SYM] THEN
      ASM_SIMP_TAC [GSYM REAL_LT_RDIV_EQ]]]);;


let Z_INITIAL_VALUE_THM = prove(
     ` !f. exterior_circle (roc f) ==>
       f(0) = lim (at_infinity)  (\z. z_transform f z) `,

ONCE_REWRITE_TAC[EQ_SYM_EQ] THEN 
MESON_TAC[Z_INITIAL_VALUE_THEOREM;
          TENDSTO_LIM;TRIVIAL_LIMIT_AT_INFINITY]);;

(*======================================================*)
(*                 MISC Lemmas                          *)
(*======================================================*)

let POWER_SERIES_SUMMABLE_WEAK = prove
 (`!s a. summable s a 
 ==> (!z. norm z < &1 ==> summable s (\i. a i * z pow i))`,
  IMP_REWRITE_TAC [AB_LEMMA] THEN REPEAT STRIP_TAC THEN
  EXISTS_TAC `&1` THEN REAL_ARITH_TAC);;

let Z_ROC_FUN_SUMMABLE Z_ROC_FUN_SUMMABLE = prove
 (`!f z. ~(z = Cx (&0)) /\ norm (inv z) < &1 /\ 
          summable (from 0) f  ==> z IN roc f`,
  REPEAT STRIP_TAC THEN
  SIMP_TAC [roc; complex_div; GSYM COMPLEX_POW_INV; IN_ELIM_THM] THEN
  MP_TAC
    (ISPECL [`((from 0):num->bool)`; `(f:num->complex)`]
       POWER_SERIES_SUMMABLE_WEAK) THEN
  ASM_SIMP_TAC []);;

(*======================================================*)
(*               Application SC Power Convertor         *)
(*======================================================*)

let sc_voltage_doubler = define `
    sc_voltage_doubler Q1 Q3 Qout Cp Cout Vin Vout <=> 
(!n. Q1(n) = Cp *(Vout(n) - Vin(n)) /\
     Q3(n) = Cp * Vin (n) /\
     Qout(n) = Cout * Vout (n) /\ 
     Q1(n - 1) = Cp * Vin (n - 1) /\
     Q3(n - 1) = Cp *(Vout(n - 1) - Vin (n-1)) /\
     Qout(n - 1) = Cout * Vout (n - 1)) `;;


let sc_parameters_constraints = define `
    sc_parameters_constraints Vin Vout Cp Rout Cout Ts z <=>
    (!n. Cx (&4) * Cp * (Vin n + Vin (n - 1)) =
       (Cx (&2) * Cp + Cout - Cx Ts / (Cx (&2) * Cx Rout)) *
        Vout n + (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) *
        Vout (n - 1)) /\ 
    (~(Cx (&2) * Cp - Cout = Cx (&0))) /\ (~(Cp = Cx(&0))) /\
     ~(Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout) = Cx (&0)) /\ 
     ~(Cx Rout = Cx(&0)) /\
     ~((Cx (&2) * Cp + Cout - Cx Ts / (Cx (&2) * Cx Rout)) +
         z cpow --Cx (&1) * (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) = Cx (&0)) /\
     ~(z_transform Vin z = Cx (&0)) /\  ~(Cx (&0) = z)  `;;


let t_charge_transfer = define `
    t_charge_transfer Q1 Q3 Qout Vout Rout Ts n <=>
 Cx(&2) * (Q1(n - 1) - Q1(n) + Q3(n) - Q3(n - 1)) + 
             (Qout(n - 1) - Qout(n)) =
     Cx(Ts) / Cx(&2) * ((Vout(n - 1) - Vout(n)) / Cx(Rout))`;;


let SC_LEMMA_1 = prove
 (`!Q1 Q3 Qout Cp Vin Vout n.
   sc_voltage_doubler Q1 Q3 Qout Cp Cout Vin Vout ==>
   t_charge_transfer Q1 Q3 Qout Vout Rout Ts n = 
  (Cx (&2) * Cp *
  (Cx(&2) * Vin (n - 1) + Cx(&2)* Vin(n) - Vout n - Vout (n - 1)) 
  + Cout * (Vout (n - 1) -  Vout n) =
  Cx Ts / Cx (&2) * (Vout (n - 1) - Vout n) / Cx Rout)`,
  REWRITE_TAC [t_charge_transfer; sc_voltage_doubler] THEN SIMP_TAC [] THEN
  REPEAT STRIP_TAC THEN CONV_TAC COMPLEX_FIELD);;


let SC_LEMMA_11 = prove
 (`!Q1 Q3 Qout Cp Vin Vout n.
   sc_voltage_doubler Q1 Q3 Qout Cp Cout Vin Vout ==>
   t_charge_transfer Q1 Q3 Qout Vout Rout Ts n =
  
(Cx (&4) * Cp  * (Vin n + Vin (n - 1))  =
  
 (Cx (&2) * Cp  + Cout - Cx Ts / (Cx (&2) * Cx Rout))* Vout n +
 
 (Cx (&2) * Cp  - Cout  +  Cx Ts / (Cx (&2) * Cx Rout)) * Vout (n - 1))`,
  SIMP_TAC [SC_LEMMA_1] THEN CONV_TAC COMPLEX_FIELD);;


let SC_LEMMA_2 =
COMPLEX_FIELD ` ! a b (d:complex). ~(a = Cx(&0)) ==>  a * (b / a + d) = b + d*a`;;


let SUMMABLE_Z_POW_LEMMA = prove 
(`!f z a. ~(Cx(&0) = z) ==>
        (summable (from 0) (\n. a* f (n - 1) / z pow n) <=>
     z_tr_summable_cpow (\n. a * f (n - 1)) z)`,

REPEAT STRIP_TAC THEN 
SIMP_TAC[complex_div] THEN 
IMP_REWRITE_TAC[GSYM CPOW_POW] THEN 
SIMP_TAC[GSYM CPOW_NEG] THEN 
SIMP_TAC[z_tr_summable_cpow;COMPLEX_MUL_ASSOC]);;


let SUMMABLE_Z_POW_LEMMA1 = prove 
(`!f z a. ~(Cx(&0) = z) ==>
        (summable (from 0) (\n.  f (n - 1) / z pow n) <=>
     z_tr_summable_cpow (\n. f (n - 1)) z)`,

REPEAT STRIP_TAC THEN 
SIMP_TAC[complex_div] THEN 
IMP_REWRITE_TAC[GSYM CPOW_POW] THEN 
SIMP_TAC[GSYM CPOW_NEG] THEN 
SIMP_TAC[z_tr_summable_cpow;COMPLEX_MUL_ASSOC]);;

let SUMMABLE_Z_POW_LEMMA2 = prove 
(`!f z a. ~(Cx(&0) = z) ==>
        (summable (from 0) (\n. (a* f (n - 1)) / z pow n) <=>
     z_tr_summable_cpow (\n. a * f (n - 1)) z)`,

REPEAT STRIP_TAC THEN 
SIMP_TAC[complex_div] THEN 
IMP_REWRITE_TAC[GSYM CPOW_POW] THEN 
SIMP_TAC[GSYM CPOW_NEG] THEN 
SIMP_TAC[z_tr_summable_cpow;COMPLEX_MUL_ASSOC]);;


let SUMMABLE_Z_POW_LEMMA3 = prove 
(`!f z a. ~(Cx(&0) = z) ==>
        (summable (from 0) (\n. (a* f (n)) / z pow n) <=>
     z_tr_summable_cpow (\n. a * f (n)) z)`,

REPEAT STRIP_TAC THEN 
SIMP_TAC[complex_div] THEN 
IMP_REWRITE_TAC[GSYM CPOW_POW] THEN 
SIMP_TAC[GSYM CPOW_NEG] THEN 
SIMP_TAC[z_tr_summable_cpow;COMPLEX_MUL_ASSOC]);;


let SUMMABLE_Z_POW_LEMMA4 = prove 
(`!f z a. ~(Cx(&0) = z) ==>
        (summable (from 0) (\n. (f (n) + g (n-1)) / z pow n) <=>
     z_tr_summable_cpow (\n. f (n) + g (n-1)) z)`,

REPEAT STRIP_TAC THEN 
SIMP_TAC[complex_div] THEN 
IMP_REWRITE_TAC[GSYM CPOW_POW] THEN 
SIMP_TAC[GSYM CPOW_NEG] THEN 
SIMP_TAC[z_tr_summable_cpow;COMPLEX_MUL_ASSOC]);;


let ZTR_SUMMABLE_ADD  = prove 
    (`!f g. z_tr_summable_cpow f z /\ z_tr_summable_cpow (\n. g (n-1)) z ==> 
       z_tr_summable_cpow (\n. f n + g (n-1)) z`,
SIMP_TAC[ z_tr_summable_cpow;COMPLEX_ADD_RDISTRIB] THEN
ONCE_REWRITE_TAC[COMPLEX_FIELD`(\n. f n * z cpow --Cx (&n) + g (n - 1) * z cpow --Cx (&n)) = 
(\n. (\n. f n * z cpow --Cx (&n)) n + (\n.g (n - 1) * z cpow --Cx (&n)) n)`] THEN 
IMP_REWRITE_TAC[SUMMABLE_ADD]);;



let SC_VOLTAGE_DOUBLER_TF_WEAK = prove
 (`!Q1 Q3 Qout n Vin Vout Cp Cout Ts Rout z.
   sc_voltage_doubler Q1 Q3 Qout Cp Cout Vin Vout /\
   (t_charge_transfer Q1 Q3 Qout Vout Rout Ts n) /\
   sc_parameters_constraints Vin Vout Cp Rout Cout Ts z /\
   z IN roc Vin /\ 
   z IN roc Vout /\
   is_causal_signal Vin /\ 
   is_causal_signal Vout /\
   z IN roc (\n. Vin (n - 1)) /\

   z IN roc (\n. (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) * Vout (n - 1)) /\
   z IN roc (\n. (Cx (&2) * Cp + Cout + Cx Ts / (Cx (&2) * Cx Rout)) * Vout (n - 1)) /\
   z IN roc (\n. Vout (n - 1)) /\ 
   z IN roc (\n. (Cx (&2) * Cp + Cout - Cx Ts / (Cx (&2) * Cx Rout)) * Vout n) /\
   z IN roc (\n. Vin n + Vin (n - 1)) ==>

transfer_function Vin Vout z = (Cx(&4) * Cp * (Cx(&1) + z cpow --Cx(&1))) /
                 ((Cx(&2)* Cp - Cout  + Cx (Ts) / (Cx(&2)* Cx Rout)) *
                 ((Cx(&2)* Cp + Cout  - Cx (Ts) / (Cx(&2)* Cx Rout))/
                  (Cx(&2)* Cp - Cout  + Cx (Ts) / (Cx(&2)* Cx Rout)) +
                   z cpow --Cx(&1)))`,
  REWRITE_TAC [sc_parameters_constraints] THEN REPEAT STRIP_TAC THEN
  IMP_REWRITE_TAC [SC_LEMMA_2] THEN REWRITE_TAC [transfer_function] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD
       `! a b c d. ~(d = Cx(&0)) /\  ~(b = Cx(&0)) ==> ( (a / b = c / d ) = (a*d = b*c ))`] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `z_transform Vin z * Cx (&4) * Cp * (Cx (&1) + z cpow --Cx (&1)) = 
Cx (&4) * Cp * z_transform Vin z * (Cx (&1) + z cpow --Cx (&1))`] THEN
  ASSERT_TAC
    `Cx (&4) * Cp * z_transform Vin z * (Cx (&1) + z cpow --Cx (&1)) = 
 (z_transform (\n. (Cx (&4) * Cp) * (\n. Vin (n)   +  Vin (n - 1))n) z)` THENL
   [IMP_REWRITE_TAC [Z_TRANSFORM_CONST_MUL] THEN
    ASSERT_TAC
      `(\n. Vin n + Vin (n - 1)) = 
(\n. (\n. Vin n) n  + (\n. Vin (n - 1))n)` THENL
     [SIMP_TAC []; ALL_TAC] THEN
    ONCE_ASM_REWRITE_TAC [] THEN IMP_REWRITE_TAC [Z_TRANSFORM_ADD] THEN
    SIMP_TAC [ETA_AX] THEN ASM_SIMP_TAC [] THEN
    IMP_REWRITE_TAC [ZTRANSFORM_TIME_DELAY] THEN SIMP_TAC [CPOW_NEG] THEN
    IMP_REWRITE_TAC [CPOW_POW] THEN ASM_SIMP_TAC [EQ_SYM_EQ] THEN
    SIMP_TAC [COMPLEX_POW_1] THEN SIMP_TAC [complex_div] THEN
    CONV_TAC COMPLEX_FIELD;
    ASM_SIMP_TAC [] THEN
    ASSERT_TAC
      `z_transform Vout z *
 ((Cx (&2) * Cp + Cout - Cx Ts / (Cx (&2) * Cx Rout)) +
  z cpow --Cx (&1) * (Cx (&2) * Cp - Cout + Cx Ts /(Cx (&2) * Cx Rout))) = 
(z_transform (\n. (\n. (Cx (&2) * Cp + Cout - Cx Ts / (Cx (&2) * Cx Rout)) *  Vout n ) n + 
(\n. (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) * 
Vout (n-1)) n ) z)` THENL
     [ALL_TAC;
      ASM_REWRITE_TAC [] THEN
      UNDISCH_TAC
        `!n. Cx (&4) * Cp * (Vin n + Vin (n - 1)) =
          (Cx (&2) * Cp + Cout - Cx Ts / (Cx (&2) * Cx Rout)) * Vout n +
          (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) * Vout (n - 1)` THEN
      SIMP_TAC [SC_LEMMA_1] THEN SIMP_TAC [COMPLEX_MUL_ASSOC; EQ_SYM_EQ]] THEN
    IMP_REWRITE_TAC [Z_TRANSFORM_ADD] THEN
    ASSERT_TAC
      `(\n. (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) * Vout (n - 1)) =  (\n. (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) * (\n. Vout (n - 1))n)` THENL
     [SIMP_TAC []; ALL_TAC] THEN
    ONCE_ASM_REWRITE_TAC [] THEN IMP_REWRITE_TAC [Z_TRANSFORM_CONST_MUL] THEN
    ASSERT_TAC
      `(\n. (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) * Vout (n - 1)) =  (\n. (Cx (&2) * Cp - Cout + Cx Ts / (Cx (&2) * Cx Rout)) * (\n. Vout (n - 1))n)` THENL
     [SIMP_TAC []; ALL_TAC] THEN
    ONCE_ASM_REWRITE_TAC [] THEN IMP_REWRITE_TAC [ZTRANSFORM_TIME_DELAY] THEN
    ASSERT_TAC
      `z_transform Vout z / z pow 1 = 
            z_transform Vout z * z cpow --Cx(&1)` THENL
     [IMP_REWRITE_TAC [complex_div; CPOW_NEG; GSYM CPOW_POW];
      POP_ASSUM MP_TAC THEN SIMP_TAC [] THEN DISCH_TAC THEN SIMP_TAC [ETA_AX] THEN
      CONV_TAC COMPLEX_FIELD]]);;


let SC_VOLTAGE_TF_LEMMA = prove 
(` !z. sc_parameters_constraints Vin Vout Cp Rout Cout Ts z ==> ~(z = Cx(&0))`,

MESON_TAC[sc_parameters_constraints]);;


let SC_VOLTAGE_DOUBLER_TRANSFER_FUNCTION = prove
 (`!Q1 Q3 Qout n Vin Vout Cp Cout Ts Rout z.
   sc_voltage_doubler Q1 Q3 Qout Cp Cout Vin Vout /\
   (t_charge_transfer Q1 Q3 Qout Vout Rout Ts n) /\
   sc_parameters_constraints Vin Vout Cp Rout Cout Ts z /\
   z IN roc Vin /\ 
   z IN roc Vout /\
   is_causal_signal Vin /\ 
   is_causal_signal Vout  ==>

transfer_function Vin Vout z = (Cx(&4) * Cp * (Cx(&1) + z cpow --Cx(&1))) /
                 ((Cx(&2)* Cp - Cout  + Cx (Ts) / (Cx(&2)* Cx Rout)) *
                 ((Cx(&2)* Cp + Cout  - Cx (Ts) / (Cx(&2)* Cx Rout))/
                  (Cx(&2)* Cp - Cout  + Cx (Ts) / (Cx(&2)* Cx Rout)) +
                   z cpow --Cx(&1)))`,
  REPEAT STRIP_TAC THEN MATCH_MP_TAC SC_VOLTAGE_DOUBLER_TF_WEAK THEN
  ASM_SIMP_TAC [] THEN ASM_SIMP_TAC [roc; IN_ELIM_THM; SC_VOLTAGE_TF_LEMMA] THEN
  IMP_REWRITE_TAC
    [SUMMABLE_Z_POW_LEMMA; SC_VOLTAGE_TF_LEMMA;
     COMPLEX_FIELD `~(Cx (&0) = z) =   ~(z = Cx(&0))`] THEN
  EXISTS_TAC `Q1:num->complex` THEN EXISTS_TAC `Q3:num->complex` THEN
  EXISTS_TAC `Qout:num->complex` THEN EXISTS_TAC `n:num` THEN ASM_SIMP_TAC [] THEN
  REPEAT STRIP_TAC THENL
   [IMP_REWRITE_TAC [SUMMABLE_Z_POW_LEMMA1] THEN
    ASM_SIMP_TAC [SC_VOLTAGE_TF_LEMMA] THEN
    ONCE_REWRITE_TAC
      [COMPLEX_FIELD `!n. Vin (n - 1 ) = Cx(&1) * Vin (n - 1)`] THEN
    IMP_REWRITE_TAC [Z_SUMMABLE_CMUL_DIFF; ZTR_ROC_EQ];
    IMP_REWRITE_TAC
      [SUMMABLE_Z_POW_LEMMA2; SC_VOLTAGE_TF_LEMMA;
       COMPLEX_FIELD `~(Cx (&0) = z) =   ~(z = Cx(&0))`] THEN
    IMP_REWRITE_TAC [Z_SUMMABLE_CMUL_DIFF; ZTR_ROC_EQ];
    IMP_REWRITE_TAC
      [SUMMABLE_Z_POW_LEMMA2; SC_VOLTAGE_TF_LEMMA;
       COMPLEX_FIELD `~(Cx (&0) = z) =   ~(z = Cx(&0))`] THEN
    IMP_REWRITE_TAC [Z_SUMMABLE_CMUL_DIFF; ZTR_ROC_EQ];
    IMP_REWRITE_TAC [SUMMABLE_Z_POW_LEMMA1] THEN
    ASM_SIMP_TAC [SC_VOLTAGE_TF_LEMMA] THEN
    ONCE_REWRITE_TAC
      [COMPLEX_FIELD `!n. Vin (n - 1 ) = Cx(&1) * Vin (n - 1)`] THEN
    IMP_REWRITE_TAC [Z_SUMMABLE_CMUL_DIFF; ZTR_ROC_EQ];
    IMP_REWRITE_TAC
      [SUMMABLE_Z_POW_LEMMA3; SUMMABLE_COMPLEX_LMUL; SC_VOLTAGE_TF_LEMMA;
       COMPLEX_FIELD `~(Cx (&0) = z) =   ~(z = Cx(&0))`] THEN
    REWRITE_TAC [GSYM ROC_Z_SUMMABLE_CPOW] THEN
    IMP_REWRITE_TAC [ROC_COMPLEX_LMUL_CPOW; ZTR_ROC_ROC];
    IMP_REWRITE_TAC
      [SUMMABLE_Z_POW_LEMMA4; SUMMABLE_COMPLEX_LMUL; SC_VOLTAGE_TF_LEMMA;
       COMPLEX_FIELD `~(Cx (&0) = z) =   ~(z = Cx(&0))`] THEN
    IMP_REWRITE_TAC [ZTR_SUMMABLE_ADD; ZTR_ROC_ROC; GSYM ROC_Z_SUMMABLE_CPOW] THEN
    ASM_SIMP_TAC [roc; IN_ELIM_THM; SC_VOLTAGE_TF_LEMMA] THEN
    IMP_REWRITE_TAC [SUMMABLE_Z_POW_LEMMA1] THEN
    ASM_SIMP_TAC [SC_VOLTAGE_TF_LEMMA] THEN
    ONCE_REWRITE_TAC
      [COMPLEX_FIELD `!n. Vin (n - 1 ) = Cx(&1) * Vin (n - 1)`] THEN
    IMP_REWRITE_TAC [Z_SUMMABLE_CMUL_DIFF; ZTR_ROC_EQ]] THEN
  ASM_MESON_TAC [SC_VOLTAGE_TF_LEMMA]);;



let SC_GAIN_LEMMA = prove
 (`~(Cx (&2) * Cp - Cout = Cx (&0))/\ ~(Cp = Cx(&0))  ==> 

 (Cx (&4) * Cp * (Cx (&1) + Cx(&1) cpow --Cx (&1))) /
    ((Cx (&2) * Cp - Cout + (Cx(&0)) / (Cx (&2) * Cx Rout)) *
    ((Cx (&2) * Cp + Cout - Cx(&0) / (Cx (&2) * Cx Rout)) /
     (Cx (&2) * Cp - Cout + Cx(&0) / (Cx (&2) * Cx Rout)) +
     Cx(&1) cpow --Cx (&1))) = Cx(&2)`,
  SIMP_TAC [CPOW_1; COMPLEX_FIELD `Cx(&0) / t = Cx(&0)`] THEN
  SIMP_TAC [COMPLEX_ADD_RID; COMPLEX_FIELD `x - Cx(&0) = x`] THEN
  IMP_REWRITE_TAC
    [COMPLEX_FIELD `~(c = Cx(&0)) ==> ( b / c  + d) = (b + c*d) / c`] THEN
  IMP_REWRITE_TAC [COMPLEX_FIELD `~(a = Cx(&0)) ==> a * b / a = b`] THEN
  SIMP_TAC [COMPLEX_MUL_RID] THEN
  SIMP_TAC
    [COMPLEX_FIELD
       `((Cx (&2) * Cp + Cout) + Cx (&2) * Cp - Cout) =  (Cx (&2) * Cp + Cx (&2) * Cp )`] THEN
  SIMP_TAC [GSYM CX_ADD; CX_MUL] THEN SIMP_TAC [REAL_FIELD `&1 + &1 = &2`] THEN
  CONV_TAC COMPLEX_FIELD);;


let VOLTAGE_DOUBLER_GAIN = prove
 (`!Q1 Q3 Qout n Cp Rout Cout Vin Vout.
    is_causal_signal Vin /\
    is_causal_signal Vout /\
    summable (from 0) Vin /\
    summable (from 0) Vout /\
    sc_voltage_doubler Q1 Q3 Qout Cp Cout Vin Vout /\
    t_charge_transfer Q1 Q3 Qout Vout Rout (&0) n /\
    sc_parameters_constraints Vin Vout Cp Rout Cout (&0) (Cx (&1)) ==>
    transfer_function Vin Vout (Cx(&1)) = Cx(&2)`,
  IMP_REWRITE_TAC [SC_VOLTAGE_DOUBLER_TRANSFER_FUNCTION] THEN
  SIMP_TAC [roc; IN_ELIM_THM; COMPLEX_POW_ONE; COMPLEX_DIV_1; ETA_AX] THEN
  REPEAT STRIP_TAC THENL
   [MATCH_MP_TAC SC_GAIN_LEMMA THEN POP_ASSUM MP_TAC THEN
    SIMP_TAC [sc_parameters_constraints];
    ASM_MESON_TAC [COMPLEX_FIELD `~(Cx(&1) = Cx(&0))`]]);;


(*======================================================*)
(*                 IIR Filter                           *)
(*======================================================*)

let iir_model = new_definition `
    iir_model x y alist blist M N n  = lccde x y alist blist M N n`;;

let IIR_TRANSFER_FUNCTION = prove
 (`!z. z IN LCCDE_ROC x y M alst /\
       is_causal_lccde x y /\
       (!n. iir_model x y alst blst M N n) /\
       ~(z_transform x z = Cx (&0)) ==> 
        z_transform y z / z_transform x z =
        vsum (0..N) (\n. z cpow --Cx (&n) * EL n blst) /
        (Cx (&1) - vsum (1..M) (\n. z cpow --Cx (&n) * EL n alst))`,
  SIMP_TAC [LCCDE_TRANSFER_FUNCTION; iir_model]);;


let FRQ_LEMMA = prove
 (`! N blist a.  vsum (a..N) (\n. (j * csin (Cx (&n) * w)) * EL n blst) = 
    j* vsum (a..N) (\n. csin (Cx (&n) * w) * EL n blst)`,
  REWRITE_TAC[j] THEN SIMP_TAC [GSYM COMPLEX_MUL_ASSOC] THEN
  ASSERT_TAC
    `!j. (\n. j * csin (Cx (&n) * w) * EL n blst) = 
(\n. j * (\n. csin (Cx (&n) * w) * EL n blst) n)` THENL
   [SIMP_TAC []; ALL_TAC] THEN
  ONCE_ASM_REWRITE_TAC [] THEN
  IMP_REWRITE_TAC [VSUM_COMPLEX_LMUL; FINITE_NUMSEG]);;


let FREQ_RESPONSE_IIR = prove
 (`!x y N blst M w alst. cexp (j * w) IN LCCDE_ROC x y M alst /\
 is_causal_lccde x y /\
 (!n. iir_model x y alst blst M N n) /\
 ~(z_transform x (cexp (j * w)) = Cx (&0)) ==>

let H = transfer_function x y (cexp(j*w)) and

num_real =     vsum (0..N) (\n. ccos (Cx (&n) * w) * EL n blst) and 
num_im   =  j * vsum (0..N) (\n. (csin (Cx (&n) * w)) * EL n blst) and 

denom_real = Cx (&1) -
   	    vsum (1..M) (\n. ccos (Cx (&n) * w) * EL n alst) and 
denom_im = j * vsum (1..M) (\n. csin (Cx (&n) * w) * EL n alst) in

H = Cx (norm (num_real - num_im) / norm (denom_real + denom_im)) *
 cexp (j * Cx (Arg ((num_real - num_im) / (denom_real + denom_im))))`,
  REWRITE_TAC [iir_model] THEN REPEAT STRIP_TAC THEN REPEAT LET_TAC THEN
  ABBREV_TAC
    `ttt = Cx (norm (num_real - num_im) / norm (denom_real + denom_im)) *
      cexp (j * Cx (Arg ((num_real - num_im) / (denom_real + denom_im))))` THEN
  UNDISCH_TAC `transfer_function x y (cexp (j * w)) = H` THEN
  SIMP_TAC [EQ_SYM_EQ] THEN DISCH_TAC THEN ONCE_REWRITE_TAC [EQ_SYM_EQ] THEN
  REWRITE_TAC [transfer_function] THEN
  MP_TAC (SPECL [`cexp(j*w)`] IIR_TRANSFER_FUNCTION) THEN
  ASM_SIMP_TAC [iir_model] THEN REPEAT STRIP_TAC THEN SIMP_TAC [CPOW_NEG] THEN
  IMP_REWRITE_TAC [CPOW_POW; CEXP_NZ] THEN REWRITE_TAC [GSYM CEXP_N] THEN
  SIMP_TAC [COMPLEX_FIELD `!j. (Cx (&n) * j * w) = 
j * ((Cx (&n)* w))`] THEN
  SIMP_TAC [GSYM CEXP_NEG] THEN
  SIMP_TAC [COMPLEX_FIELD `!j. (--(j * Cx (&n) * w)) = 
(j * (--Cx (&n) * w))`] THEN
  SIMP_TAC [REWRITE_RULE [GSYM j] CEXP_EULER] THEN
  SIMP_TAC [COMPLEX_FIELD `(--Cx (&n) * w) = 
(--(Cx (&n) * w))`] THEN
  SIMP_TAC [CCOS_NEG; CSIN_NEG] THEN
  SIMP_TAC [COMPLEX_FIELD `!j. a + j * --b = a - j*b`] THEN
  SIMP_TAC [COMPLEX_SUB_RDISTRIB] THEN SIMP_TAC [VSUM_SUB; FINITE_NUMSEG] THEN
  SIMP_TAC [FRQ_LEMMA] THEN
  SIMP_TAC [COMPLEX_FIELD `a - (b - c) = a - b + (c:complex)`] THEN
  ASM_SIMP_TAC [] THEN
  UNDISCH_TAC
    `Cx (norm (num_real - num_im) / norm (denom_real + denom_im)) *
      cexp (j * Cx (Arg ((num_real - num_im) / (denom_real + denom_im)))) =
      ttt` THEN
  SIMP_TAC [GSYM (REWRITE_RULE [GSYM j] ARG); GSYM COMPLEX_NORM_DIV]);;

